# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

import ess.isissans.data  # noqa: F401
from ess import isissans as isis
from ess import sans
from ess.sans.types import (
    BeamCenter,
    CorrectForGravity,
    DetectorPositionOffset,
    Filename,
    Incident,
    IntensityQ,
    IntensityQxQy,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    QBins,
    QxBins,
    QyBins,
    SampleRun,
    Transmission,
    UncertaintyBroadcastMode,
    WavelengthBins,
)


def make_params() -> dict:
    params = {
        **isis.default_parameters(),
        sans.types.DirectBeamFilename: isis.data.zoom_tutorial_direct_beam(),
        isis.CalibrationFilename: isis.data.zoom_tutorial_calibration(),
        Filename[sans.types.SampleRun]: isis.data.zoom_tutorial_sample_run(),
        Filename[sans.types.EmptyBeamRun]: isis.data.zoom_tutorial_empty_beam_run(),
        isis.general.SampleOffset: sc.vector([0.0, 0.0, 0.11], unit='m'),
        DetectorPositionOffset[SampleRun]: sc.vector([0.0, 0.0, 0.5], unit='m'),
    }

    params[NeXusMonitorName[Incident]] = 'monitor3'
    params[NeXusMonitorName[Transmission]] = 'monitor5'

    params[WavelengthBins] = sc.geomspace(
        'wavelength', start=1.75, stop=16.5, num=141, unit='angstrom'
    )

    params[QBins] = sc.geomspace(
        dim='Q', start=0.004, stop=0.8, num=141, unit='1/angstrom'
    )

    params[NonBackgroundWavelengthRange] = sc.array(
        dims=['wavelength'], values=[0.7, 17.1], unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[sans.ReturnEvents] = False
    return params


@pytest.fixture
def pipeline():
    wf = isis.zoom.ZoomTutorialWorkflow()
    wf.insert(isis.io.transmission_from_background_run)
    wf.insert(isis.io.transmission_from_sample_run)
    for key, param in make_params().items():
        wf[key] = param
    return wf


def test_can_create_pipeline(pipeline):
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, isis.data.zoom_tutorial_mask_filenames()
    )
    pipeline.get(IntensityQ[SampleRun])


def test_pipeline_can_compute_IofQ(pipeline):
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, isis.data.zoom_tutorial_mask_filenames()
    )
    result = pipeline.compute(IntensityQ[SampleRun])
    assert result.dims == ('Q',)


def test_pipeline_can_compute_IofQxQy(pipeline):
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, isis.data.zoom_tutorial_mask_filenames()
    )
    pipeline[QxBins] = sc.linspace(
        dim='Qx', start=-0.5, stop=0.5, num=101, unit='1/angstrom'
    )
    pipeline[QyBins] = sc.linspace(
        dim='Qy', start=-0.8, stop=0.8, num=101, unit='1/angstrom'
    )

    result = pipeline.compute(IntensityQxQy[SampleRun])
    assert result.dims == ('Qy', 'Qx')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from functools import lru_cache

import pytest
import sciline
import scipp as sc

import ess.isissans.data  # noqa: F401
from ess import isissans as isis
from ess import sans
from ess.isissans import MonitorOffset, SampleOffset, sans2d
from ess.sans.types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BeamCenter,
    CorrectedDetector,
    CorrectForGravity,
    DimsToKeep,
    DirectBeam,
    DirectBeamFilename,
    EmptyBeamRun,
    EmptyDetector,
    Filename,
    Incident,
    IntensityQ,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    Numerator,
    QBins,
    RawDetector,
    ReturnEvents,
    SampleRun,
    SolidAngle,
    Transmission,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
    WavelengthMask,
)


def make_params() -> dict:
    params = isis.default_parameters()
    params[WavelengthBins] = sc.linspace(
        'wavelength', start=2.0, stop=16.0, num=141, unit='angstrom'
    )
    params[WavelengthMask] = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[True]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[2.21, 2.59], unit='angstrom'
            )
        },
    )
    params[sans2d.LowCountThreshold] = sc.scalar(100.0, unit='counts')

    params[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.55, num=141, unit='1/angstrom'
    )
    params[DirectBeamFilename] = isis.data.sans2d_tutorial_direct_beam()
    params[Filename[SampleRun]] = isis.data.sans2d_tutorial_sample_run()
    params[Filename[BackgroundRun]] = isis.data.sans2d_tutorial_background_run()
    params[Filename[EmptyBeamRun]] = isis.data.sans2d_tutorial_empty_beam_run()

    params[NeXusMonitorName[Incident]] = 'monitor2'
    params[NeXusMonitorName[Transmission]] = 'monitor4'
    params[SampleOffset] = sc.vector([0.0, 0.0, 0.053], unit='m')
    params[MonitorOffset[Transmission]] = sc.vector([0.0, 0.0, -6.719], unit='m')

    params[NonBackgroundWavelengthRange] = sc.array(
        dims=['wavelength'], values=[0.7, 17.1], unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[ReturnEvents] = False
    params[DimsToKeep] = ()
    params[BeamCenter] = sc.vector([0, 0, 0], unit='m')

    return params


# Especially the I(Q) beam-cetner finder calls this many times, so we cache it for
# faster test runtime.
cached_load = lru_cache(maxsize=None)(isis.io.load_tutorial_run)


# Wrapper adding type-hints back to the cached function
def cached_load_tutorial_run(
    filename: Filename[SampleRun],
) -> isis.io.LoadedFileContents[SampleRun]:
    return cached_load(filename)


@pytest.fixture
def pipeline():
    wf = isis.sans2d.Sans2dTutorialWorkflow()
    wf.insert(isis.io.transmission_from_background_run)
    wf.insert(isis.io.transmission_from_sample_run)
    wf.insert(cached_load_tutorial_run)
    for key, param in make_params().items():
        wf[key] = param
    return wf


def test_can_create_pipeline(pipeline):
    pipeline.get(IntensityQ[SampleRun])


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pipeline_can_compute_background_subtracted_IofQ(pipeline, uncertainties):
    pipeline[UncertaintyBroadcastMode] = uncertainties
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_pipeline_can_compute_background_subtracted_IofQ_in_wavelength_bands(pipeline):
    pipeline[WavelengthBands] = sc.linspace(
        'wavelength', start=2.0, stop=16.0, num=11, unit='angstrom'
    )
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('band', 'Q')
    assert result.sizes['band'] == 10


def test_pipeline_wavelength_bands_is_optional(pipeline):
    pipeline[BeamCenter] = sans.beam_center_from_center_of_mass(pipeline)
    noband = pipeline.compute(BackgroundSubtractedIofQ)
    assert pipeline.compute(WavelengthBands) is None
    band = sc.linspace('wavelength', 2.0, 16.0, num=2, unit='angstrom')
    pipeline[WavelengthBands] = band
    assert sc.identical(band, pipeline.compute(WavelengthBands))
    withband = pipeline.compute(BackgroundSubtractedIofQ)
    assert sc.identical(noband, withband)


def test_workflow_is_deterministic(pipeline):
    pipeline[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    pipeline[BeamCenter] = sans.beam_center_from_center_of_mass(pipeline)
    # This is Sciline's default scheduler, but we want to be explicit here
    scheduler = sciline.scheduler.DaskScheduler()
    graph = pipeline.get(IntensityQ[SampleRun], scheduler=scheduler)
    reference = graph.compute().data
    result = graph.compute().data
    assert sc.identical(sc.values(result), sc.values(reference))


def test_pipeline_raises_VariancesError_if_normalization_errors_not_dropped(pipeline):
    pipeline[NonBackgroundWavelengthRange] = (
        None  # Make sure we raise in iofq_denominator
    )
    pipeline[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.fail
    with pytest.raises(sc.VariancesError):
        pipeline.compute(BackgroundSubtractedIofQ)


def test_uncertainty_broadcast_mode_drop_yields_smaller_variances(pipeline):
    # Errors with the full range have some NaNs or infs
    pipeline[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.5, num=141, unit='1/angstrom'
    )
    pipeline[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    drop = pipeline.compute(IntensityQ[SampleRun]).data
    pipeline[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    upper_bound = pipeline.compute(IntensityQ[SampleRun]).data
    assert sc.all(sc.variances(drop) < sc.variances(upper_bound)).value


def test_pipeline_can_visualize_background_subtracted_IofQ(pipeline):
    pipeline.visualize(BackgroundSubtractedIofQ)


def test_pipeline_can_compute_intermediate_results(pipeline):
    result = pipeline.compute(SolidAngle[SampleRun])
    assert result.dims == ('spectrum',)


def pixel_dependent_direct_beam(
    filename: DirectBeamFilename, shape: EmptyDetector[SampleRun]
) -> DirectBeam:
    direct_beam = isis.io.load_tutorial_direct_beam(filename)
    sizes = {'spectrum': shape.sizes['spectrum'], **direct_beam.sizes}
    return DirectBeam(direct_beam.broadcast(sizes=sizes).copy())


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pixel_dependent_direct_beam_is_supported(pipeline, uncertainties):
    pipeline[UncertaintyBroadcastMode] = uncertainties
    pipeline.insert(pixel_dependent_direct_beam)
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


MANTID_BEAM_CENTER = sc.vector([0.09288, -0.08195, 0], unit='m')


def test_beam_center_from_center_of_mass_is_close_to_verified_result(pipeline):
    center = sans.beam_center_from_center_of_mass(pipeline)
    # This is the result obtained from Mantid, using the full IofQ
    # calculation. The difference is about 3 mm in X or Y, probably due to a bias
    # introduced by the sample holder, which the center-of-mass approach cannot ignore.
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(3e-3, unit='m'))


def test_beam_center_from_center_of_mass_independent_of_set_beam_center(pipeline):
    pipeline[BeamCenter] = sc.vector([0.1, -0.1, 0], unit='m')
    center = sans.beam_center_from_center_of_mass(pipeline)
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(3e-3, unit='m'))


def test_beam_center_finder_without_direct_beam_reproduces_verified_result(pipeline):
    pipeline[DirectBeam] = None
    center = sans.beam_center_finder.beam_center_from_iofq(
        workflow=pipeline, q_bins=sc.linspace('Q', 0.02, 0.3, 71, unit='1/angstrom')
    )
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(2e-3, unit='m'))


def test_beam_center_can_get_closer_to_verified_result_with_low_counts_mask(pipeline):
    def low_counts_mask(
        sample: RawDetector[SampleRun],
        low_counts_threshold: sans2d.LowCountThreshold,
    ) -> sans2d.SampleHolderMask:
        return sans2d.SampleHolderMask(sample.hist().data < low_counts_threshold)

    pipeline[sans2d.LowCountThreshold] = sc.scalar(80.0, unit='counts')
    pipeline.insert(low_counts_mask)  # replaces sans2d.sample_holder_mask
    pipeline[DirectBeam] = None
    q_bins = sc.linspace('Q', 0.02, 0.3, 71, unit='1/angstrom')
    center = sans.beam_center_finder.beam_center_from_iofq(
        workflow=pipeline, q_bins=q_bins
    )
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(5e-4, unit='m'))


def test_beam_center_finder_works_with_direct_beam(pipeline):
    q_bins = sc.linspace('Q', 0.02, 0.3, 71, unit='1/angstrom')
    center_with_direct_beam = sans.beam_center_finder.beam_center_from_iofq(
        workflow=pipeline, q_bins=q_bins
    )
    assert sc.allclose(
        center_with_direct_beam, MANTID_BEAM_CENTER, atol=sc.scalar(2e-3, unit='m')
    )


def test_beam_center_finder_independent_of_set_beam_center(pipeline):
    pipeline[BeamCenter] = sc.vector([0.1, -0.1, 0], unit='m')
    q_bins = sc.linspace('Q', 0.02, 0.3, 71, unit='1/angstrom')
    center_with_direct_beam = sans.beam_center_finder.beam_center_from_iofq(
        workflow=pipeline, q_bins=q_bins
    )
    assert sc.allclose(
        center_with_direct_beam, MANTID_BEAM_CENTER, atol=sc.scalar(2e-3, unit='m')
    )


def test_beam_center_finder_works_with_pixel_dependent_direct_beam(pipeline):
    q_bins = sc.linspace('Q', 0.02, 0.3, 71, unit='1/angstrom')
    center_pixel_independent_direct_beam = (
        sans.beam_center_finder.beam_center_from_iofq(workflow=pipeline, q_bins=q_bins)
    )

    direct_beam = pipeline.compute(DirectBeam)
    pixel_dependent_direct_beam = direct_beam.broadcast(
        sizes={
            'spectrum': pipeline.compute(CorrectedDetector[SampleRun, Numerator]).sizes[
                'spectrum'
            ],
            'wavelength': direct_beam.sizes['wavelength'],
        }
    ).copy()

    pipeline[DirectBeam] = pixel_dependent_direct_beam

    center = sans.beam_center_finder.beam_center_from_iofq(
        workflow=pipeline, q_bins=q_bins
    )
    assert sc.identical(center, center_pixel_independent_direct_beam)


def test_workflow_runs_without_gravity_if_beam_center_is_provided(pipeline):
    pipeline[CorrectForGravity] = False
    da = pipeline.compute(RawDetector[SampleRun])
    del da.coords['gravity']
    pipeline[RawDetector[SampleRun]] = da
    pipeline[BeamCenter] = MANTID_BEAM_CENTER
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from pathlib import Path

from ess.reduce.data import make_registry
from ess.sans.types import (
    BackgroundRun,
    DirectBeamFilename,
    Filename,
    PixelMaskFilename,
    SampleRun,
    TransmissionRun,
)

_registry = make_registry(
    'ess/loki',
    files={
        # Files from LoKI@Larmor detector test experiment.
        # Original files are available at:
        # https://project.esss.dk/nextcloud/index.php/apps/files/?dir=/Data/LOKI_detector_test/nexus/2022-06-24_calibrated_nexus_files  # noqa: E501
        #
        # Background run 1 (no sample, sample holder/can only, no transmission monitor)
        '60248-2022-02-28_2215.nxs': 'md5:d9f17b95274a0fc6468df7e39df5bf03',
        # Sample run 1 (sample + sample holder/can, no transmission monitor in beam)
        '60250-2022-02-28_2215.nxs': 'md5:6a519ceaacbae702a6d08241e86799b1',
        # Sample run 2 (sample + sample holder/can, no transmission monitor in beam)
        '60339-2022-02-28_2215.nxs': 'md5:03c86f6389566326bb0cbbd80b8f8c4f',
        # Background transmission run (sample holder/can + transmission monitor)
        '60392-2022-02-28_2215.nxs': 'md5:9ecc1a9a2c05a880144afb299fc11042',
        # Background run 2 (no sample, sample holder/can only, no transmission monitor)
        '60393-2022-02-28_2215.nxs': 'md5:bf550d0ba29931f11b7450144f658652',
        # Sample transmission run (sample + sample holder/can + transmission monitor)
        '60394-2022-02-28_2215.nxs': 'md5:c40f38a62337d86957af925296c4c615',
        # ISIS polymer sample run
        "60395-2022-02-28_2215.nxs": "md5:d4c8ac05d0a015f2808089255e8ead3c",
        # AgBeh sample run
        "60387-2022-02-28_2215.nxs": "md5:157b937f00a5da133481b22b78ec7fa1",
        # AgBeh transmission run
        "60386-2022-02-28_2215.nxs": "md5:58ef33133e86e0026ee36f1a24deb464",
        # Porous silica sample run
        "60385-2022-02-28_2215.nxs": "md5:21275cb80c146d6c424bea81142b9a76",
        # Porous silica transmission run
        "60384-2022-02-28_2215.nxs": "md5:af04240efd3a245280f2d9f4846c6076",
        # deut-SDS sample run
        "60389-2022-02-28_2215.nxs": "md5:3126ec7a670ac603c6a5f4c756ddc5b7",
        # deut-SDS transmission run
        "60388-2022-02-28_2215.nxs": "md5:72737d0e796a5b7bb4241dd8157c5905",
        # Analytical model for the I(Q) of the Poly-Gauss sample
        'PolyGauss_I0-50_Rg-60.h5': 'md5:f5d60d9c2286cb197b8cd4dc82db3d7e',
        # XML file for the pixel mask
        'mask_new_July2022.xml': 'md5:421b6dc9db74126ffbc5d88164d017b0',
        # Direct beam from LoKI@Larmor detector test experiment
        'direct-beam-loki-all-pixels.h5': "md5:b85d7b486b312c5bb2a31d2bb6314f69",
        # CODA file with a single event per detector bank (for testing purposes)
        'loki-coda-one-event.hdf': 'md5:ab9dbef793fad2fca96210c3b55d60ce',
        # Small CODA file for testing purposes (data from only 30 pulses)
        'loki_999999_00009928_small.hdf': 'md5:d2f293c293cc15d90fb588afa1dcb096',
        # Wavelength lookup table without choppers
        'loki-wavelength-lookup-table-no-choppers.h5': 'md5:3fb133962af6554a9999a753c1724ed4',  # noqa: E501
    },
    version='2',
)


def loki_tutorial_sample_run_60250() -> Filename[SampleRun]:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return Filename[SampleRun](_registry.get_path('60250-2022-02-28_2215.nxs'))


def loki_tutorial_sample_run_60339() -> Filename[SampleRun]:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return Filename[SampleRun](_registry.get_path('60339-2022-02-28_2215.nxs'))


def loki_tutorial_background_run_60248() -> Filename[BackgroundRun]:
    """Background run with sample holder/can only, no transmission monitor."""
    return Filename[BackgroundRun](_registry.get_path('60248-2022-02-28_2215.nxs'))


def loki_tutorial_background_run_60393() -> Filename[BackgroundRun]:
    """Background run with sample holder/can only, no transmission monitor."""
    return Filename[BackgroundRun](_registry.get_path('60393-2022-02-28_2215.nxs'))


def loki_tutorial_sample_transmission_run() -> Filename[TransmissionRun[SampleRun]]:
    """Sample transmission run (sample + sample holder/can + transmission monitor)."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path('60394-2022-02-28_2215.nxs')
    )


def loki_tutorial_run_60392() -> Filename[TransmissionRun[BackgroundRun]]:
    """Background transmission run (sample holder/can + transmission monitor), also
    used as empty beam run."""
    return Filename[TransmissionRun[BackgroundRun]](
        _registry.get_path('60392-2022-02-28_2215.nxs')
    )


def loki_tutorial_isis_polymer_sample_run() -> Filename[SampleRun]:
    """Sample run with ISIS polymer sample."""
    return Filename[SampleRun](_registry.get_path("60395-2022-02-28_2215.nxs"))


def loki_tutorial_isis_polymer_transmission_run() -> (
    Filename[TransmissionRun[SampleRun]]
):
    """Transmission run for ISIS polymer run."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path("60394-2022-02-28_2215.nxs")
    )


def loki_tutorial_agbeh_sample_run() -> Filename[SampleRun]:
    """Sample run with AgBeh sample."""
    return Filename[SampleRun](_registry.get_path("60387-2022-02-28_2215.nxs"))


def loki_tutorial_agbeh_transmission_run() -> Filename[TransmissionRun[SampleRun]]:
    """Transmission run for AgBeh run."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path("60386-2022-02-28_2215.nxs")
    )


def loki_tutorial_porous_silica_sample_run() -> Filename[SampleRun]:
    """Sample run with Porous silica sample."""
    return Filename[SampleRun](_registry.get_path("60385-2022-02-28_2215.nxs"))


def loki_tutorial_porous_silica_transmission_run() -> (
    Filename[TransmissionRun[SampleRun]]
):
    """Transmission run for Porous silica run."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path("60384-2022-02-28_2215.nxs")
    )


def loki_tutorial_deut_sds_sample_run() -> Filename[SampleRun]:
    """Sample run with deut-SDS sample."""
    return Filename[SampleRun](_registry.get_path("60389-2022-02-28_2215.nxs"))


def loki_tutorial_deut_sds_transmission_run() -> Filename[TransmissionRun[SampleRun]]:
    """Transmission run for deut-SDS run."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path("60388-2022-02-28_2215.nxs")
    )


def loki_tutorial_mask_filenames() -> list[PixelMaskFilename]:
    """List of pixel mask filenames for the LoKI@Larmor detector test experiment."""
    return [
        PixelMaskFilename(_registry.get_path('mask_new_July2022.xml')),
    ]


def loki_tutorial_poly_gauss_I0() -> Path:
    """Analytical model for the I(Q) of the Poly-Gauss sample."""
    return Path(_registry.get_path('PolyGauss_I0-50_Rg-60.h5'))


def loki_tutorial_direct_beam_all_pixels() -> DirectBeamFilename:
    """File containing direct beam function computed using the direct beam iterations
    notebook, summing all pixels."""
    return DirectBeamFilename(_registry.get_path('direct-beam-loki-all-pixels.h5'))


def loki_lookup_table_no_choppers() -> Path:
    """
    LoKI wavelength lookup table without choppers.
    This file is used to convert the neutron arrival time to wavelength.

    This table was computed using `Create a wavelength lookup table for LoKI
    <../../user-guide/loki/loki-make-wavelength-lookup-table.ipynb>`_
    with ``NumberOfSimulatedNeutrons = 5_000_000``.
    """
    return _registry.get_path("loki-wavelength-lookup-table-no-choppers.h5")


def loki_coda_file_one_event() -> Path:
    """
    LoKI CODA file with a single event per detector bank (for testing purposes).
    The file was created from a CODA file 'loki_999999_00009928.hdf' but the event lists
    were replaced by lists with a single event to reduce the size of the file.
    The workflow should still be able to run, it will just produce results with NaNs
    everywhere.
    """
    return _registry.get_path("loki-coda-one-event.hdf")


def loki_coda_file_small() -> Path:
    """
    Small LoKI CODA file for testing purposes (data from only 30 pulses).
    """
    return _registry.get_path("loki_999999_00009928_small.hdf")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc
from scipp.scipy.interpolate import interp1d

from ess import loki, sans
from ess.sans.types import (
    BeamCenter,
    DimsToKeep,
    QBins,
    WavelengthBands,
    WavelengthBins,
)


def _get_I0(qbins: sc.Variable) -> sc.Variable:
    Iq_theory = sc.io.load_hdf5(loki.data.loki_tutorial_poly_gauss_I0())
    f = interp1d(Iq_theory, 'Q')
    return f(sc.midpoints(qbins)).data[0]


def test_can_compute_direct_beam_for_all_pixels(larmor_workflow):
    n_wavelength_bands = 10
    pipeline = larmor_workflow()
    edges = pipeline.compute(WavelengthBins)
    pipeline[WavelengthBands] = sc.linspace(
        'wavelength', edges.min(), edges.max(), n_wavelength_bands + 1
    )
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    I0 = _get_I0(qbins=pipeline.compute(QBins))

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('Q',)
    assert iofq_bands.dims == ('band', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands


def test_can_compute_direct_beam_with_overlapping_wavelength_bands(larmor_workflow):
    n_wavelength_bands = 10
    # Bands have double the width
    pipeline = larmor_workflow()
    edges = pipeline.compute(WavelengthBins)
    edges = sc.linspace('band', edges.min(), edges.max(), n_wavelength_bands + 2)
    pipeline[WavelengthBands] = sc.concat(
        [edges[:-2], edges[2::]], dim='wavelength'
    ).transpose()

    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    I0 = _get_I0(qbins=pipeline.compute(QBins))

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('Q',)
    assert iofq_bands.dims == ('band', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands


def test_can_compute_direct_beam_per_layer(larmor_workflow):
    n_wavelength_bands = 10
    pipeline = larmor_workflow()
    edges = pipeline.compute(WavelengthBins)
    pipeline[WavelengthBands] = sc.linspace(
        'wavelength', edges.min(), edges.max(), n_wavelength_bands + 1
    )
    pipeline[DimsToKeep] = ['layer']
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    I0 = _get_I0(qbins=pipeline.compute(QBins))

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('layer', 'Q')
    assert iofq_bands.dims == ('band', 'layer', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert iofq_bands.sizes['layer'] == 4
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands
    assert direct_beam_function.sizes['layer'] == 4


def test_can_compute_direct_beam_per_layer_and_straw(larmor_workflow):
    n_wavelength_bands = 10
    pipeline = larmor_workflow()
    edges = pipeline.compute(WavelengthBins)
    pipeline[WavelengthBands] = sc.linspace(
        'wavelength', edges.min(), edges.max(), n_wavelength_bands + 1
    )
    pipeline[DimsToKeep] = ('layer', 'straw')
    pipeline[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    I0 = _get_I0(qbins=pipeline.compute(QBins))

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('layer', 'straw', 'Q')
    assert iofq_bands.dims == ('band', 'layer', 'straw', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert iofq_bands.sizes['layer'] == 4
    assert iofq_bands.sizes['straw'] == 7
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands
    assert direct_beam_function.sizes['layer'] == 4
    assert direct_beam_function.sizes['straw'] == 7



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401, I

import importlib.metadata

from . import (
    beam_center_finder,
    common,
    conversions,
    i_of_q,
    io,
    masking,
    normalization,
)
from .beam_center_finder import beam_center_from_center_of_mass
from .direct_beam import direct_beam
from .types import BackgroundSubtractedIofQ, IntensityQ, ReturnEvents, SampleRun
from .workflow import (
    SansWorkflow,
    providers,
    with_background_runs,
    with_banks,
    with_pixel_mask_filenames,
    with_sample_runs,
)

try:
    __version__ = importlib.metadata.version("esssans")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = [
    'BackgroundSubtractedIofQ',
    'IntensityQ',
    'ReturnEvents',
    'SampleRun',
    'SansWorkflow',
    'beam_center_finder',
    'beam_center_from_center_of_mass',
    'common',
    'conversions',
    'direct_beam',
    'i_of_q',
    'io',
    'masking',
    'normalization',
    'providers',
    'with_background_runs',
    'with_banks',
    'with_pixel_mask_filenames',
    'with_sample_runs',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import scipp as sc
import scippnexus as snx
from scipp.core import concepts

from ess.reduce.uncertainty import UncertaintyBroadcastMode, broadcast_uncertainties

from .types import (
    CleanDirectBeam,
    CorrectedDetector,
    CorrectedMonitor,
    Denominator,
    DetectorMasks,
    DetectorPixelShape,
    DetectorTerm,
    EmptyBeamRun,
    EmptyDetector,
    Incident,
    IntensityQ,
    IntensityQxQy,
    IofQPart,
    MonitorTerm,
    NeXusTransformation,
    NormalizedQ,
    NormalizedQxQy,
    Numerator,
    Position,
    ProcessedWavelengthBands,
    ReducedQ,
    ReducedQxQy,
    ReturnEvents,
    RunType,
    SolidAngle,
    Transmission,
    TransmissionFraction,
    TransmissionRun,
    WavelengthBands,
    WavelengthBins,
)


def solid_angle(
    data: EmptyDetector[RunType],
    pixel_shape: DetectorPixelShape[RunType],
    transform: NeXusTransformation[snx.NXdetector, RunType],
    sample_position: Position[snx.NXsample, RunType],
) -> SolidAngle[RunType]:
    """
    Solid angle for cylindrical pixels.

    Note that the approximation is valid when the distance from sample
    to pixel is much larger than the length or the radius of the pixels.

    Parameters
    ----------
    data:
        The DataArray that contains the positions of the detector pixels
        and the position of the sample in the coords.

    pixel_shape:
        Contains the description of the detector pixel shape.

    transform:
        Transformation from the local coordinate system of the detector
        to the coordinate system of the sample.

    Returns
    -------
    :
        The solid angle of the detector pixels, as viewed from the sample position.
    """
    face_1_center, face_1_edge, face_2_center = (
        pixel_shape['vertices']['vertex', i] for i in range(3)
    )
    cylinder_axis = transform.value * face_2_center - transform.value * face_1_center
    radius = sc.norm(face_1_center - face_1_edge)
    length = sc.norm(cylinder_axis)

    omega = _approximate_solid_angle_for_cylinder_shaped_pixel_of_detector(
        pixel_position=data.coords['position'] - sample_position,
        cylinder_axis=cylinder_axis,
        radius=radius,
        length=length,
    )
    return SolidAngle[RunType](
        concepts.rewrap_reduced_data(
            prototype=data, data=omega, dim=set(data.dims) - set(omega.dims)
        )
    )


def mask_solid_angle(
    solid_angle: SolidAngle[RunType],
    masks: DetectorMasks,
) -> CorrectedDetector[RunType, Denominator]:
    return CorrectedDetector[RunType, Denominator](solid_angle.assign_masks(masks))


def _approximate_solid_angle_for_cylinder_shaped_pixel_of_detector(
    pixel_position: sc.Variable,
    cylinder_axis: sc.Variable,
    radius: sc.Variable,
    length: sc.Variable,
):
    r"""Computes solid angle of a detector pixel under the assumption that the
    distance to the detector pixel from the sample is large compared to the radius
    and to the length of the piece of the detector cylinder that makes up the pixel.

    The cylinder is approximated by a rectangle element contained in the cylinder.
    The normal of the rectangle is selected to
    1. be orthogonal to the cylinder axis and
    2. "maximally parallel" to the pixel position vector.
    This occurs when the normal :math:`n`, the cylinder axis :math:`c` and the
    pixel position :math:`r` fall in a shared plane such that
    .. math::
        r = (n\cdot r)n + (c \cdot r)c.

    From the above relationship we get the scalar product of the position
    vector and the normal:

    .. math::
        |\hat{r}\cdot n| = \sqrt{\frac{r \cdot r - (c \cdot r^2)}{||r||^2}}

    The solid angle contribution from the detector element is approximated as
    .. math::
        \Delta\Omega =  A \frac{|\hat{r}\cdot n|}{||r||^2}
    where :math:`A = 2 R L` is the area of the rectangular element
    and :math:`\hat{r}` is the normalized pixel position vector.
    """
    norm_pp = sc.norm(pixel_position)
    norm_ca = sc.norm(cylinder_axis)
    cosalpha = sc.sqrt(
        1 - (sc.dot(pixel_position, cylinder_axis) / (norm_pp * norm_ca)) ** 2
    )
    return (2 * radius * length) * cosalpha / norm_pp**2


def transmission_fraction(
    sample_incident_monitor: CorrectedMonitor[TransmissionRun[RunType], Incident],
    sample_transmission_monitor: CorrectedMonitor[
        TransmissionRun[RunType], Transmission
    ],
    direct_incident_monitor: CorrectedMonitor[EmptyBeamRun, Incident],
    direct_transmission_monitor: CorrectedMonitor[EmptyBeamRun, Transmission],
) -> TransmissionFraction[RunType]:
    """
    Approximation based on equations in
    `CalculateTransmission <https://docs.mantidproject.org/v4.0.0/algorithms/CalculateTransmission-v1.html>`_
    documentation:
    ``(sample_transmission_monitor / direct_transmission_monitor) * (direct_incident_monitor / sample_incident_monitor)``

    This is equivalent to ``mantid.CalculateTransmission`` without fitting.
    Inputs should be wavelength-dependent.

    Parameters
    ----------
    sample_incident_monitor:
        The incident monitor data for the sample (transmission) run.
    sample_transmission_monitor:
        The transmission monitor data for the sample (transmission) run.
    direct_incident_monitor:
        The incident monitor data for the direct beam run.
    direct_transmission_monitor:
        The transmission monitor data for the direct beam run.

    Returns
    -------
    :
        The transmission fraction computed from the monitor counts.
    """  # noqa: E501
    frac = (sample_transmission_monitor / direct_transmission_monitor) * (
        direct_incident_monitor / sample_incident_monitor
    )
    return TransmissionFraction[RunType](frac)


def norm_monitor_term(
    incident_monitor: CorrectedMonitor[RunType, Incident],
    transmission_fraction: TransmissionFraction[RunType],
) -> MonitorTerm[RunType]:
    """
    Compute the monitor-dependent contribution to the denominator term of I(Q).

    This is basically ``incident_monitor * transmission_fraction``.

    Keeping the monitor term separate from the detector term allows us to compute
    the latter only once when repeatedly processing chunks of events in streamed data
    processing, e.g., for live data reduction.

    Parameters
    ----------
    incident_monitor:
        The incident monitor data (depends on wavelength).
    transmission_fraction:
        The transmission fraction (depends on wavelength).

    Returns
    -------
    :
        Monitor-dependent factor of the normalization term for the SANS I(Q).
    """
    out = incident_monitor * transmission_fraction
    # Convert wavelength coordinate to midpoints for future histogramming
    out.coords['wavelength'] = sc.midpoints(out.coords['wavelength'])
    return MonitorTerm[RunType](out)


def norm_detector_term_denominator(
    solid_angle: CorrectedDetector[RunType, Denominator],
    direct_beam: CleanDirectBeam,
    uncertainties: UncertaintyBroadcastMode,
) -> DetectorTerm[RunType, Denominator]:
    """
    Compute the detector-dependent contribution to the denominator term of I(Q).

    This is basically ``solid_angle * direct_beam``.
    If the direct beam is not supplied, it is assumed to be 1.

    Keeping the monitor term separate from the detector term allows us to compute the
    the latter only once when repeatedly processing chunks of events in streamed data
    processing, e.g., for live data reduction.

    Because the multiplication between the pixel-dependent solid angle and the
    (potentially) pixel-independent direct beam constitutes a broadcast operation which
    would introduce correlations, variances of the direct beam are dropped or replaced
    by an upper-bound estimation, depending on the configured mode.

    Parameters
    ----------
    solid_angle:
        The solid angle of the detector pixels, as viewed from the sample position.
    direct_beam:
        The direct beam function (depends on wavelength).
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`ess.reduce.uncertainty.UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        Detector-dependent factor of the normalization term for the SANS I(Q).
    """
    # Make wavelength the inner dim
    dims = list(direct_beam.dims)
    dims.remove('wavelength')
    dims.append('wavelength')
    direct_beam = direct_beam.transpose(dims)
    out = solid_angle * broadcast_uncertainties(
        direct_beam, prototype=solid_angle, mode=uncertainties
    )
    # Convert wavelength coordinate to midpoints for future histogramming
    out.coords['wavelength'] = sc.midpoints(out.coords['wavelength'])
    return DetectorTerm[RunType, Denominator](out)


def norm_detector_term_numerator(
    detector: CorrectedDetector[RunType, Numerator],
) -> DetectorTerm[RunType, Numerator]:
    """
    A dummy provider to convert CorrectedDetector into DetectorTerm.
    This is added instead of having the masking operation directly return the
    DetectorTerm, as it is more consistent that Denominator and Numerator have a
    CorrectedDetector step. We also do not make :func:`compute_Q` accept
    the CorrectedDetector directly, because this would lead to complications when we
    want to swap it out for :func:`compute_Qxy`.
    """
    return DetectorTerm[RunType, Numerator](detector)


def process_wavelength_bands(
    wavelength_bands: WavelengthBands,
    wavelength_bins: WavelengthBins,
) -> ProcessedWavelengthBands:
    """
    Perform some checks and potential reshaping on the wavelength bands.

    The wavelength bands must be either one- or two-dimensional.
    If the wavelength bands are defined as a one-dimensional array, convert them to a
    two-dimensional array with start and end wavelengths.

    The final bands must have a size of 2 in the wavelength dimension, defining a start
    and an end wavelength.
    """
    if wavelength_bands is None:
        wavelength_bands = sc.concat(
            [wavelength_bins.min(), wavelength_bins.max()], dim='wavelength'
        )
    if wavelength_bands.ndim == 1:
        wavelength_bands = sc.concat(
            [wavelength_bands[:-1], wavelength_bands[1:]], dim='x'
        ).rename(x='wavelength', wavelength='band')
    if wavelength_bands.ndim != 2:
        raise ValueError(
            'Wavelength_bands must be one- or two-dimensional, '
            f'got {wavelength_bands.ndim}.'
        )
    if wavelength_bands.sizes['wavelength'] != 2:
        raise ValueError(
            'Wavelength_bands must have a size of 2 in the wavelength dimension, '
            'defining a start and an end wavelength, '
            f'got {wavelength_bands.sizes["wavelength"]}.'
        )
    return wavelength_bands


def _normalize(
    numerator: sc.DataArray,
    denominator: sc.DataArray,
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
) -> sc.DataArray:
    """
    Perform normalization of counts as a function of Q to obtain I(Q) or I(Qx, Qy).
    division.

    In a SANS experiment, the scattering cross section :math:`I(Q)` is defined as
    (`Heenan et al. 1997 <https://doi.org/10.1107/S0021889897002173>`_):

    .. math::

       I(Q) = \\frac{\\partial\\Sigma{Q}}{\\partial\\Omega} = \\frac{A_{H} \\Sigma_{R,\\lambda\\subset Q} C(R, \\lambda)}{A_{M} t \\Sigma_{R,\\lambda\\subset Q}M(\\lambda)T(\\lambda)D(\\lambda)\\Omega(R)}

    where :math:`A_{H}` is the area of a mask (which avoids saturating the detector)
    placed between the monitor of area :math:`A_{M}` and the main detector.
    :math:`\\Omega` is the detector solid angle, and :math:`C` is the count rate on the
    main detector, which depends on the position :math:`R` and the wavelength.
    :math:`t` is the sample thickness, :math:`M` represents the incident monitor count
    rate for the sample run, and :math:`T` is known as the transmission fraction.

    Note that the incident monitor used to compute the transmission fraction is not
    necessarily the same as :math:`M`, as the transmission fraction is usually computed
    from a separate 'transmission' run (in the 'sample' run, the transmission monitor is
    commonly moved out of the beam path, to avoid polluting the sample detector signal).

    Finally, :math:`D` is the 'direct beam function', and is defined as

    .. math::

       D(\\lambda) = \\frac{\\eta(\\lambda)}{\\eta_{M}(\\lambda)} \\frac{A_{H}}{A_{M}}

    where :math:`\\eta` and :math:`\\eta_{M}` are the detector and monitor
    efficiencies, respectively.

    Hence, in order to normalize the main detector counts :math:`C`, we need compute the
    transmission fraction :math:`T(\\lambda)`, the direct beam function
    :math:`D(\\lambda)` and the solid angle :math:`\\Omega(R)`.

    The denominator is then simply:
    :math:`M_{\\lambda} T_{\\lambda} D_{\\lambda} \\Omega_{R}`.

    Parameters
    ----------
    numerator:
        The data whose counts will be divided by the denominator. This can either be
        event or dense (histogrammed) data.
    denominator:
        The divisor for the normalization operation. This cannot be event data, it must
        contain histogrammed data.
    return_events:
        Whether to return the result as event data or histogrammed data.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`ess.reduce.uncertainty.UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        Normalized I(Q) or I(Qx, Qy).
    """  # noqa: E501
    if return_events and numerator.bins is not None:
        # Naive event-mode normalization is not correct if norm-term has variances.
        # See https://doi.org/10.3233/JNR-220049 for context.
        denominator = broadcast_uncertainties(
            denominator, prototype=numerator, mode=uncertainties
        )
    elif numerator.bins is not None:
        numerator = numerator.hist()
    numerator /= denominator.drop_coords(
        [name for name in denominator.coords if name not in denominator.dims]
    )
    return numerator


def _do_reduce(da: sc.DataArray) -> sc.DataArray:
    wav = 'wavelength'
    if da.sizes[wav] == 1:  # Can avoid costly event-data da.bins.concat
        return da.squeeze(wav)
    return da.sum(wav) if da.bins is None else da.bins.concat(wav)


def _reduce(part: sc.DataArray, /, *, bands: ProcessedWavelengthBands) -> sc.DataArray:
    """
    Reduce data by summing or concatenating along the wavelength dimension.

    Parameters
    ----------
    data:
        Numerator or denominator data to be reduced.
    wavelength_bands:
        Defines bands in wavelength that can be used to separate different wavelength
        ranges that contribute to different regions in Q space. Note that this needs to
        be defined, so if all wavelengths should be used, this should simply be a start
        and end edges that encompass the entire wavelength range.

    Returns
    -------
    :
        Q-dependent data, ready for normalization.
    """
    wav = 'wavelength'
    if part.bins is not None:
        # If in event mode the desired wavelength binning has not been applied, we need
        # it for splitting by bands, or restricting the range in case of a single band.
        part = part.bin(wavelength=sc.sort(bands.flatten(to=wav), wav))
    parts = [
        _do_reduce(part[wav, wav_range[0] : wav_range[1]])
        for wav_range in sc.collapse(bands, keep=wav).values()
    ]
    band_dim = (set(bands.dims) - {'wavelength'}).pop()
    reduced = parts[0] if len(parts) == 1 else sc.concat(parts, band_dim)
    return reduced.assign_coords(wavelength=bands.squeeze())


def reduce_q(
    data: NormalizedQ[RunType, IofQPart],
    bands: ProcessedWavelengthBands,
) -> ReducedQ[RunType, IofQPart]:
    return ReducedQ[RunType, IofQPart](_reduce(data, bands=bands))


def reduce_qxy(
    data: NormalizedQxQy[RunType, IofQPart],
    bands: ProcessedWavelengthBands,
) -> ReducedQxQy[RunType, IofQPart]:
    return ReducedQxQy[RunType, IofQPart](_reduce(data, bands=bands))


def normalize_q(
    numerator: ReducedQ[RunType, Numerator],
    denominator: ReducedQ[RunType, Denominator],
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
) -> IntensityQ[RunType]:
    return IntensityQ[RunType](
        _normalize(
            numerator=numerator,
            denominator=denominator,
            return_events=return_events,
            uncertainties=uncertainties,
        )
    )


def normalize_qxy(
    numerator: ReducedQxQy[RunType, Numerator],
    denominator: ReducedQxQy[RunType, Denominator],
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
) -> IntensityQxQy[RunType]:
    return IntensityQxQy[RunType](
        _normalize(
            numerator=numerator,
            denominator=denominator,
            return_events=return_events,
            uncertainties=uncertainties,
        )
    )


reduce_q.__doc__ = _reduce.__doc__
reduce_qxy.__doc__ = _reduce.__doc__
normalize_q.__doc__ = _normalize.__doc__
normalize_qxy.__doc__ = _normalize.__doc__


providers = (
    transmission_fraction,
    norm_monitor_term,
    norm_detector_term_denominator,
    norm_detector_term_numerator,
    reduce_q,
    reduce_qxy,
    normalize_q,
    normalize_qxy,
    process_wavelength_bands,
    solid_angle,
    mask_solid_angle,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Detector diagnostics for LOKI."""

from __future__ import annotations

import ipywidgets as ipw
import matplotlib.pyplot as plt
import plopp as pp
import scipp as sc


def _add_missing_coordinates(da: sc.DataArray) -> sc.DataArray:
    return da.assign_coords(
        {dim: sc.arange(dim=dim, start=1, stop=sz + 1) for dim, sz in da.sizes.items()}
    )


def _slice_or_sum(
    da: sc.DataGroup, layer_ind: int, layer_sum: bool, straw_ind: int, straw_sum: bool
) -> sc.DataGroup:
    out = da.copy(deep=False)
    if layer_sum:
        out = out.sum('layer')
    else:
        out = out["layer", layer_ind - 1]
    if straw_sum:
        out = out.sum('straw')
    else:
        out = out["straw", straw_ind - 1]
    return out


class LokiBankViewer(ipw.VBox):
    def __init__(self, data: sc.DataGroup, figsize=(12, 9), **kwargs):
        """Widget to view LOKI detector banks.

        Parameters
        ----------
        data:
            DataGroup containing LOKI detector banks.
        figsize:
            Size of the figures in inches (width, height).
        kwargs:
            Additional arguments are forwarded to the plotting function.
        """
        self.data = sc.DataGroup(
            {k: _add_missing_coordinates(da) for k, da in data.items()}
        )
        self._lock = False

        self.layer_slider = ipw.IntSlider(
            min=1, max=4, description="Layer", style={"description_width": 'initial'}
        )
        self.layer_sum = ipw.Checkbox(
            description="Sum", value=False, indent=False, layout={"width": "initial"}
        )

        self.layer_ind_node = pp.widget_node(self.layer_slider)
        self.layer_sum_node = pp.widget_node(self.layer_sum)

        self.straw_slider = ipw.IntSlider(
            min=1, max=7, description="Straw", style={"description_width": 'initial'}
        )
        self.straw_sum = ipw.Checkbox(
            description="Sum", value=False, indent=False, layout={"width": "initial"}
        )
        self.straw_ind_node = pp.widget_node(self.straw_slider)
        self.straw_sum_node = pp.widget_node(self.straw_sum)

        self.layer_link = ipw.jslink(
            (self.layer_sum, 'value'), (self.layer_slider, 'disabled')
        )
        self.straw_link = ipw.jslink(
            (self.straw_sum, 'value'), (self.straw_slider, 'disabled')
        )

        self.slice_node = pp.Node(
            _slice_or_sum,
            da=self.data,
            layer_ind=self.layer_ind_node,
            layer_sum=self.layer_sum_node,
            straw_ind=self.straw_ind_node,
            straw_sum=self.straw_sum_node,
        )

        with plt.ioff():
            self.main_figure, axs = plt.subplots(3, 3, figsize=figsize)
            bank_figures = [plt.subplots(figsize=figsize) for _ in range(9)]

        self.subplots = []
        self.tab_figs = []
        self.nodes = []
        for i, ax in enumerate(axs.flatten()):
            bank = f"loki_detector_{i}"
            n = pp.Node(lambda da, key: da[key], da=self.slice_node, key=bank)
            self.nodes.append(n)
            self.subplots.append(
                pp.imagefigure(n, ax=ax, title=bank, cbar=True, **kwargs)
            )
            self.tab_figs.append(
                pp.imagefigure(
                    ax=bank_figures[i][1],
                    title=bank,
                    cbar=True,
                    figsize=figsize,
                    **kwargs,
                )
            )
            bank_figures[i][0].canvas.header_visible = False

        self.main_figure.canvas.header_visible = False

        # Create color map controls
        self.cmap_vmin = ipw.Text(
            description='Min:',
            layout={'width': '130px'},
            style={"description_width": 'initial'},
        )
        self.cmap_vmax = ipw.Text(
            description='Max:',
            layout={'width': '130px'},
            style={"description_width": 'initial'},
        )
        self.cmap_vmin.observe(self.update_cmin, names='value')
        self.cmap_vmax.observe(self.update_cmax, names='value')

        self.log_button = ipw.ToggleButton(
            description="log",
            layout={"width": "40px"},
            value=True if kwargs.get('norm', 'linear') == 'log' else False,
        )
        self.log_button.observe(self.toggle_log, names="value")

        layer_box = ipw.HBox(
            [self.layer_slider, self.layer_sum],
            layout={'border': '1px solid black', 'padding': '0px 10px 0px 10px'},
        )
        straw_box = ipw.HBox(
            [self.straw_slider, self.straw_sum],
            layout={'border': '1px solid black', 'padding': '0px 10px 0px 10px'},
        )
        space = ipw.HTML('<div style="width: 20px;"></div>')

        self.tabs = ipw.Tab(
            layout={'width': f'{figsize[0] * self.main_figure.get_dpi() + 40}px'}
        )
        self.tabs.children = [self.main_figure.canvas] + [
            f.fig.canvas for f in self.tab_figs
        ]
        self.tabs.titles = ["All", *(f"Bank {i}" for i in range(len(self.tab_figs)))]
        self.tabs.observe(self.update_node_routing, names='selected_index')

        super().__init__(
            [
                ipw.HBox(
                    [
                        layer_box,
                        space,
                        straw_box,
                        space,
                        ipw.Label("Colorscale:"),
                        self.cmap_vmin,
                        self.cmap_vmax,
                        self.log_button,
                    ]
                ),
                self.tabs,
            ]
        )

    def update_colormapper_norm(self) -> None:
        if self.tabs.selected_index == 0:
            figs = self.subplots
        else:
            figs = [self.tab_figs[self.tabs.selected_index - 1]]
        for f in figs:
            f.view.colormapper.norm = "log" if self.log_button.value else "linear"

    def toggle_log(self, _: dict) -> None:
        self.reset_cmap_range()
        self.update_colormapper_norm()

    def _update_min_or_max(self, change: dict, bound: str) -> None:
        try:
            new = float(change["new"])
            if self.log_button.value and new <= 0:
                return
        except ValueError:
            new = None

        if self.tabs.selected_index == 0:
            figs = self.subplots
        else:
            figs = [self.tab_figs[self.tabs.selected_index - 1]]

        for f in figs:
            if new is None:
                getattr(f.view.colormapper, f"_c{bound}").pop('user', None)
            else:
                setattr(f.view.colormapper, f"c{bound}", new)
            if not self._lock:
                f.view.colormapper.autoscale()
        if not self._lock:
            if self.tabs.selected_index == 0:
                self.main_figure.canvas.draw_idle()
            else:
                self.tab_figs[self.tabs.selected_index - 1].canvas.draw()

    def update_cmin(self, change: dict) -> None:
        self._update_min_or_max(change, 'min')

    def update_cmax(self, change: dict) -> None:
        self._update_min_or_max(change, 'max')

    def reset_cmap_range(self, _=None) -> None:
        self._lock = True
        self.cmap_vmin.value = ''
        self.cmap_vmax.value = ''
        self._lock = False

    def update_node_routing(self, change: dict) -> None:
        for n in self.nodes:
            n.views.clear()
        if change['new'] == 0:
            for f, n in zip(self.subplots, self.nodes, strict=True):
                n.add_view(f.view)
        else:
            n = self.nodes[change['new'] - 1]
            f = self.tab_figs[change['new'] - 1]
            n.add_view(f.view)
        self.slice_node.notify_children(message=None)
        self.update_colormapper_norm()
        self._lock = True
        self.update_cmin({'new': self.cmap_vmin.value})
        self._lock = False
        self.update_cmax({'new': self.cmap_vmax.value})


def _to_data_group(data: sc.DataArray | sc.DataGroup | dict) -> sc.DataGroup:
    if isinstance(data, sc.DataArray):
        data = sc.DataGroup({data.name or "data": data})
    elif isinstance(data, dict):
        data = sc.DataGroup(data)
    return data


@pp.node
def _pre_process(da: sc.DataArray, dim: str) -> sc.DataArray:
    dims = list(da.dims)
    if dim is not None:
        dims.remove(dim)
    out = da.flatten(dims=dims, to="pixel")
    sel = sc.isfinite(out.coords["position"])
    return out[sel]


class InstrumentView(ipw.VBox):
    """
    Three-dimensional visualization of the Loki instrument.
    The instrument view is capable of slicing the input data with a slider widget along
    a dimension (e.g. ``tof``) by using the ``dim`` argument.
    It will also generate toggle buttons to hide/show the different modules that make up
    the Loki detectors.

    Parameters
    ----------
    data:
        Data to visualize. The data can be a single detector module (``DataArray``),
        or a group of detector modules (``dict`` or ``DataGroup``).
        The data must contain a ``position`` coordinate.
    dim:
        Dimension to use for the slider. No slider will be shown if this is None.
    pixel_size:
        Size of the pixels.
    autoscale:
        If ``True``, the color scale will be automatically adjusted to the data as it
        gets updated. This can be somewhat expensive with many pixels, so it is set to
        ``False`` by default.
    **kwargs:
        Additional arguments are forwarded to the scatter3d figure
        (see https://scipp.github.io/plopp/generated/plopp.scatter3d.html).
    """

    def __init__(
        self,
        data: sc.DataArray | sc.DataGroup | dict,
        dim: str | None = None,
        pixel_size: float | sc.Variable | None = None,
        **kwargs,
    ):
        from plopp.widgets import SliceWidget, slice_dims

        if dim and isinstance(data, sc.DataArray) and dim in data.dims[:-1]:
            data = data.transpose([d for d in data.dims if d != dim] + [dim])

        if dim and isinstance(data, sc.DataGroup):
            data = data.copy(deep=False)
            for k, v in data.items():
                if dim in v.dims[:-1]:
                    data[k] = v.transpose([d for d in v.dims if d != dim] + [dim])

        self.data = _to_data_group(data)
        self.pre_process_nodes = {
            key: _pre_process(da, dim) for key, da in self.data.items()
        }

        self._widgets = []

        if dim is not None:
            self.slider = SliceWidget(next(iter(self.data.values())), dims=[dim])
            self.slider.controls[dim].slider.layout = {"width": "600px"}
            self.slider_node = pp.widget_node(self.slider)
            self.slice_nodes = {
                key: slice_dims(n, self.slider_node)
                for key, n in self.pre_process_nodes.items()
            }
            to_scatter = self.slice_nodes
            self._widgets.append(self.slider)
        else:
            self.slice_nodes = self.pre_process_nodes
            to_scatter = self.pre_process_nodes

        kwargs.setdefault('cbar', True)
        self.fig = pp.scatter3d(
            to_scatter,
            pos="position",
            pixel_size=1.0 * sc.Unit("cm") if pixel_size is None else pixel_size,
            **kwargs,
        )

        self._widgets.insert(0, self.fig)

        if len(self.data) > 1:
            self._add_module_control()

        super().__init__(self._widgets)

    def _add_module_control(self):
        import ipywidgets as ipw

        self.cutting_tool = self.fig.bottom_bar[0]
        self._node_backup = list(self.cutting_tool._original_nodes)
        self.artist_mapping = dict(
            zip(self.data.keys(), self.fig.artists.keys(), strict=True)
        )
        self.buttons = {
            key: ipw.ToggleButton(
                value=True, description=f"{i}", layout={"width": "initial"}
            )
            for i, key in enumerate(self.data)
        }

        self.modules_widget = ipw.HBox(
            [
                ipw.HTML(value="Detector banks: &nbsp;&nbsp;&nbsp;&nbsp;"),
                *self.buttons.values(),
            ]
        )
        for key, b in self.buttons.items():
            b.key = key
            b.observe(self._check_visibility, names="value")
        self._widgets.insert(0, self.modules_widget)

    def _check_visibility(self, _):
        active_nodes = [
            node_id
            for key, node_id in self.artist_mapping.items()
            if self.buttons[key].value
        ]
        for n in self._node_backup:
            self.fig.artists[n.id].points.visible = n.id in active_nodes
        self.cutting_tool._original_nodes = [
            n for n in self._node_backup if n.id in active_nodes
        ]
        self.cutting_tool.update_state()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Scipp contributors (https://github.com/scipp)
"""
Default parameters, providers and utility functions for the Loki workflow.
"""

import sciline
import scipp as sc
import scippnexus as snx
from scippneutron.conversion.graph import beamline, tof

from ess import sans
from ess.reduce.workflow import register_workflow
from ess.sans.io import read_xml_detector_masking
from ess.sans.parameters import typical_outputs

from ..sans.conversions import ElasticCoordTransformGraph, sans_elastic
from ..sans.types import (
    BackgroundRun,
    BeamCenter,
    CorrectForGravity,
    DetectorBankSizes,
    EmptyBeamRun,
    Filename,
    GravityVector,
    Incident,
    MonitorType,
    NeXusDetectorName,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    PixelMaskFilename,
    PixelShapePath,
    Position,
    RawDetector,
    RawMonitor,
    RunType,
    SampleRun,
    Transmission,
    TransmissionRun,
    WavelengthDetector,
    WavelengthMonitor,
)
from .workflow import loki_providers

DETECTOR_BANK_SIZES = {
    'larmor_detector': {'layer': 4, 'tube': -1, 'straw': 7, 'pixel': 512},
}


class MonitorCoordTransformGraph(sciline.Scope[RunType, dict], dict):
    """
    Coordinate transformation graph that allows to compute wavelength from tof for
    Larmor monitors (no scattering).
    """


def larmor_default_parameters() -> dict:
    return {
        DetectorBankSizes: DETECTOR_BANK_SIZES,
        NeXusMonitorName[Incident]: 'monitor_1',
        NeXusMonitorName[Transmission]: 'monitor_2',
        PixelShapePath: 'pixel_shape',
        NonBackgroundWavelengthRange: None,
    }


def _larmor_convert_to_wavelength(da: sc.DataArray, graph: dict) -> sc.DataArray:
    event_time_offset = da.bins.coords['event_time_offset']
    da = da.bins.drop_coords('event_time_offset')
    da.bins.coords['tof'] = event_time_offset
    if 'event_time_zero' in da.dims:
        da = da.bins.concat('event_time_zero')
    return da.transform_coords('wavelength', graph=graph, keep_intermediate=False)


def larmor_data_to_wavelength(
    da: RawDetector[RunType], graph: ElasticCoordTransformGraph[RunType]
) -> WavelengthDetector[RunType]:
    """
    Compute wavelength coordinate for Loki detector data at Larmor.
    This is different from the standard conversion from the GenericUnwrapWorkflow
    because the detector test was conducted as ISIS where the pulse has a different time
    structure.
    The conversion here is much simpler: the event_time_offset coordinate is directly
    renamed as time-of-flight, and a wavelength is computed from that.
    """
    return WavelengthDetector[RunType](_larmor_convert_to_wavelength(da, graph=graph))


def larmor_monitor_to_wavelength(
    da: RawMonitor[RunType, MonitorType], graph: MonitorCoordTransformGraph[RunType]
) -> WavelengthMonitor[RunType, MonitorType]:
    """
    Compute wavelength coordinate for Loki monitor data at Larmor.
    This is different from the standard conversion from the GenericUnwrapWorkflow
    because the detector test was conducted as ISIS where the pulse has a different time
    structure.
    The conversion here is much simpler: the event_time_offset coordinate is directly
    renamed as time-of-flight, and a wavelength is computed from that.
    """
    return WavelengthMonitor[RunType, MonitorType](
        _larmor_convert_to_wavelength(da, graph=graph)
    )


def larmor_detector_coord_transform_graph(
    correct_for_gravity: CorrectForGravity,
    *,
    sample_position: Position[snx.NXsample, RunType],
    source_position: Position[snx.NXsource, RunType],
    gravity: GravityVector,
) -> ElasticCoordTransformGraph[RunType]:
    graph = sans_elastic(
        correct_for_gravity=correct_for_gravity,
        sample_position=sample_position,
        source_position=source_position,
        gravity=gravity,
    )
    return ElasticCoordTransformGraph[RunType]({**graph, **tof.elastic_Q('tof')})


def larmor_monitor_coord_transform_graph(
    source_position: Position[snx.NXsource, RunType],
) -> MonitorCoordTransformGraph[RunType]:
    """
    Generate a coordinate transformation graph for SANS monitor (no scattering).
    """
    return MonitorCoordTransformGraph(
        {
            **beamline.beamline(scatter=False),
            **tof.elastic_wavelength('tof'),
            'source_position': lambda: source_position,
        }
    )


loki_at_larmor_providers = (
    larmor_data_to_wavelength,
    larmor_monitor_to_wavelength,
    larmor_detector_coord_transform_graph,
    larmor_monitor_coord_transform_graph,
    read_xml_detector_masking,
)


@register_workflow
def LokiAtLarmorWorkflow() -> sciline.Pipeline:
    """
    Workflow with default parameters for Loki test at Larmor.

    This version of the Loki workflow:

    - Uses ISIS XML files to define masks.
    - Sets a dummy sample position [0,0,0] since files do not contain this information.

    Returns
    -------
    :
        Loki workflow as a sciline.Pipeline
    """
    workflow = sans.SansWorkflow()
    for provider in loki_providers:
        workflow.insert(provider)
    for provider in loki_at_larmor_providers:
        workflow.insert(provider)
    for key, param in larmor_default_parameters().items():
        workflow[key] = param
    workflow[NeXusDetectorName] = 'larmor_detector'
    workflow.typical_outputs = typical_outputs
    return workflow


@register_workflow
def LokiAtLarmorTutorialWorkflow() -> sciline.Pipeline:
    from ess.loki import data

    workflow = LokiAtLarmorWorkflow()

    workflow[PixelMaskFilename] = list(map(str, data.loki_tutorial_mask_filenames()))
    workflow[Filename[SampleRun]] = str(data.loki_tutorial_sample_run_60339())
    workflow[Filename[BackgroundRun]] = str(data.loki_tutorial_background_run_60393())
    workflow[Filename[TransmissionRun[SampleRun]]] = str(
        data.loki_tutorial_sample_transmission_run()
    )
    workflow[Filename[TransmissionRun[BackgroundRun]]] = str(
        data.loki_tutorial_run_60392()
    )
    workflow[Filename[EmptyBeamRun]] = str(data.loki_tutorial_run_60392())
    workflow[BeamCenter] = sc.vector(value=[-0.02914868, -0.01816138, 0.0], unit='m')
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
Default parameters, providers and utility functions for the Loki workflow.
"""

import sciline
import scipp as sc
import scippnexus as snx

from ess import sans
from ess.reduce.workflow import register_workflow
from ess.sans.parameters import typical_outputs

from ..sans.types import (
    DetectorBankSizes,
    DetectorPixelShape,
    DirectBeam,
    DirectBeamFilename,
    Incident,
    LookupTableRelativeErrorThreshold,
    NeXusComponent,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    PixelShapePath,
    RunType,
    Transmission,
)

DETECTOR_BANK_SIZES = {
    'loki_detector_0': {'layer': 4, 'tube': 56, 'straw': 7, 'pixel': -1},
    'loki_detector_1': {'layer': 4, 'tube': 16, 'straw': 7, 'pixel': -1},
    'loki_detector_2': {'layer': 4, 'tube': 12, 'straw': 7, 'pixel': -1},
    'loki_detector_3': {'layer': 4, 'tube': 16, 'straw': 7, 'pixel': -1},
    'loki_detector_4': {'layer': 4, 'tube': 12, 'straw': 7, 'pixel': -1},
    'loki_detector_5': {'layer': 4, 'tube': 28, 'straw': 7, 'pixel': -1},
    'loki_detector_6': {'layer': 4, 'tube': 32, 'straw': 7, 'pixel': -1},
    'loki_detector_7': {'layer': 4, 'tube': 20, 'straw': 7, 'pixel': -1},
    'loki_detector_8': {'layer': 4, 'tube': 32, 'straw': 7, 'pixel': -1},
}


def loki_default_parameters() -> dict:
    return {
        DetectorBankSizes: DETECTOR_BANK_SIZES,
        NeXusMonitorName[Incident]: 'beam_monitor_1',
        NeXusMonitorName[Transmission]: 'beam_monitor_3',
        PixelShapePath: 'pixel_shape',
        NonBackgroundWavelengthRange: None,
        LookupTableRelativeErrorThreshold: {
            **{f'loki_detector_{i}': float('inf') for i in range(9)},
            # Monitors where renamed, beam_monitor_i is the old name
            # and beam_monitor_mi is the new.
            **{f'beam_monitor_{i}': float('inf') for i in range(5)},
            **{f'beam_monitor_m{i}': float('inf') for i in range(5)},
        },
    }


def detector_pixel_shape(
    detector: NeXusComponent[snx.NXdetector, RunType],
    pixel_shape_path: PixelShapePath,
) -> DetectorPixelShape[RunType]:
    return DetectorPixelShape[RunType](detector[pixel_shape_path])


def load_direct_beam(filename: DirectBeamFilename) -> DirectBeam:
    """Load direct beam from file."""
    return DirectBeam(sc.io.load_hdf5(filename))


loki_providers = (detector_pixel_shape, load_direct_beam)


@register_workflow
def LokiWorkflow() -> sciline.Pipeline:
    """
    Workflow with default parameters for Loki.

    Returns
    -------
    :
        Loki workflow as a sciline.Pipeline
    """
    workflow = sans.SansWorkflow()
    for provider in loki_providers:
        workflow.insert(provider)
    for key, param in loki_default_parameters().items():
        workflow[key] = param
    workflow.typical_outputs = typical_outputs
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
Default parameters, providers and utility functions for the loki workflow.
"""

from __future__ import annotations

import scipp as sc

from ess.reduce.parameter import (
    BinEdgesParameter,
    BooleanParameter,
    FilenameParameter,
    MultiFilenameParameter,
    ParamWithOptions,
    StringParameter,
    Vector2dParameter,
    parameter_registry,
)

from ..sans.types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BackgroundSubtractedIofQxy,
    BeamCenter,
    CorrectedDetector,
    CorrectForGravity,
    DirectBeam,
    DirectBeamFilename,
    EmptyBeamRun,
    Filename,
    Incident,
    IntensityQ,
    IntensityQxQy,
    NeXusDetectorName,
    NeXusMonitorName,
    Numerator,
    PixelMaskFilename,
    PixelShapePath,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    SampleRun,
    TransformationPath,
    Transmission,
    TransmissionRun,
    UncertaintyBroadcastMode,
    WavelengthBins,
    WavelengthMonitor,
)

parameter_registry[CorrectForGravity] = BooleanParameter.from_type(
    CorrectForGravity, default=False
)
parameter_registry[NeXusDetectorName] = StringParameter.from_type(NeXusDetectorName)

parameter_registry[NeXusMonitorName[Incident]] = StringParameter.from_type(
    NeXusMonitorName[Incident], default=''
)
parameter_registry[NeXusMonitorName[Transmission]] = StringParameter.from_type(
    NeXusMonitorName[Transmission], default=''
)
parameter_registry[TransformationPath] = StringParameter.from_type(
    TransformationPath, default=''
)
parameter_registry[PixelMaskFilename] = MultiFilenameParameter.from_type(
    PixelMaskFilename
)
parameter_registry[PixelShapePath] = StringParameter.from_type(
    PixelShapePath, default=''
)
# Should this be ReductionMode (EventMode/HistogramMode)?
parameter_registry[ReturnEvents] = BooleanParameter.from_type(
    ReturnEvents, default=False
)
parameter_registry[UncertaintyBroadcastMode] = ParamWithOptions.from_enum(
    UncertaintyBroadcastMode, default=UncertaintyBroadcastMode.upper_bound
)
parameter_registry[Filename[SampleRun]] = MultiFilenameParameter.from_type(
    Filename[SampleRun]
)
parameter_registry[Filename[BackgroundRun]] = MultiFilenameParameter.from_type(
    Filename[BackgroundRun]
)
parameter_registry[Filename[TransmissionRun[SampleRun]]] = FilenameParameter.from_type(
    Filename[TransmissionRun[SampleRun]]
)
parameter_registry[Filename[TransmissionRun[BackgroundRun]]] = (
    FilenameParameter.from_type(Filename[TransmissionRun[BackgroundRun]])
)
parameter_registry[Filename[EmptyBeamRun]] = FilenameParameter.from_type(
    Filename[EmptyBeamRun]
)

parameter_registry[WavelengthBins] = BinEdgesParameter(
    WavelengthBins, dim='wavelength', start=2, stop=12.0, nbins=300, log=False
)
parameter_registry[QBins] = BinEdgesParameter(
    QBins, dim='Q', start=0.1, stop=0.3, nbins=100, log=False
)
parameter_registry[QxBins] = BinEdgesParameter(
    QxBins, dim='Qx', start=-0.5, stop=0.5, nbins=100
)
parameter_registry[QyBins] = BinEdgesParameter(
    QyBins, dim='Qy', start=-0.5, stop=0.5, nbins=100
)
parameter_registry[DirectBeam] = StringParameter.from_type(
    DirectBeam, switchable=True, optional=True, default=None
)
parameter_registry[DirectBeamFilename] = FilenameParameter.from_type(
    DirectBeamFilename, switchable=True
)
parameter_registry[BeamCenter] = Vector2dParameter.from_type(
    BeamCenter, default=sc.vector([0, 0, 0], unit='m')
)

typical_outputs = (
    BackgroundSubtractedIofQ,
    BackgroundSubtractedIofQxy,
    IntensityQ[SampleRun],
    IntensityQxQy[SampleRun],
    IntensityQ[BackgroundRun],
    IntensityQxQy[BackgroundRun],
    CorrectedDetector[BackgroundRun, Numerator],
    CorrectedDetector[SampleRun, Numerator],
    WavelengthMonitor[SampleRun, Incident],
    WavelengthMonitor[SampleRun, Transmission],
    WavelengthMonitor[BackgroundRun, Incident],
    WavelengthMonitor[BackgroundRun, Transmission],
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from typing import NewType

import sciline
import scipp as sc

from ess.reduce.workflow import register_workflow
from ess.sans import SansWorkflow
from ess.sans.parameters import typical_outputs
from ess.sans.types import BeamCenter, DetectorMasks, EmptyDetector, SampleRun

from .general import default_parameters
from .io import load_tutorial_direct_beam, load_tutorial_run
from .mantidio import providers as mantid_providers

DetectorEdgeMask = NewType('DetectorEdgeMask', sc.Variable | None)
"""Detector edge mask"""

LowCountThreshold = NewType('LowCountThreshold', sc.Variable)
"""Threshold below which detector pixels should be masked
(low-counts on the edges of the detector panel, and the beam stop)"""

SampleHolderMask = NewType('SampleHolderMask', sc.Variable | None)
"""Sample holder mask"""


def detector_edge_mask(
    beam_center: BeamCenter, sample: EmptyDetector[SampleRun]
) -> DetectorEdgeMask:
    # These values were determined by hand before the beam center was available.
    # We therefore undo the shift introduced by the beam center.
    raw_pos = sample.coords['position'] + beam_center
    mask_edges = (sc.abs(raw_pos.fields.x) > sc.scalar(0.48, unit='m')) | (
        sc.abs(raw_pos.fields.y) > sc.scalar(0.45, unit='m')
    )
    return DetectorEdgeMask(mask_edges)


def sample_holder_mask(
    beam_center: BeamCenter,
    sample: EmptyDetector[SampleRun],
    low_counts_threshold: LowCountThreshold,
) -> SampleHolderMask:
    # These values were determined by hand before the beam center was available.
    # We therefore undo the shift introduced by the beam center.
    raw_pos = sample.coords['position'] + beam_center
    summed = sample.hist()
    holder_mask = (
        (summed.data < low_counts_threshold)
        & (raw_pos.fields.x > sc.scalar(0, unit='m'))
        & (raw_pos.fields.x < sc.scalar(0.42, unit='m'))
        & (raw_pos.fields.y < sc.scalar(0.05, unit='m'))
        & (raw_pos.fields.y > sc.scalar(-0.15, unit='m'))
    )
    return SampleHolderMask(holder_mask)


def to_detector_masks(
    edge_mask: DetectorEdgeMask, holder_mask: SampleHolderMask
) -> DetectorMasks:
    """Gather detector masks.

    Unlike :py:func:`ess.sans.masking.to_detector_mask`, this function does not rely on
    mapping using a table of mask filenames. Instead it directly returns a dictionary
    if multiple masks.

    Parameters
    ----------
    edge_mask:
        Mask for detector edges.
    holder_mask:
        Mask for sample holder.
    """
    masks = {}
    if edge_mask is not None:
        masks['edges'] = edge_mask
    if holder_mask is not None:
        masks['holder_mask'] = holder_mask
    return DetectorMasks(masks)


providers = (detector_edge_mask, sample_holder_mask, to_detector_masks)


@register_workflow
def Sans2dWorkflow() -> sciline.Pipeline:
    """Create Sans2d workflow with default parameters."""
    from . import providers as isis_providers

    # Note that the actual NeXus loading in this workflow will not be used for the
    # ISIS files, the providers inserted below will replace those steps.
    workflow = SansWorkflow()
    for provider in isis_providers + mantid_providers + providers:
        workflow.insert(provider)
    for key, param in default_parameters().items():
        workflow[key] = param
    workflow.typical_outputs = typical_outputs
    return workflow


@register_workflow
def Sans2dTutorialWorkflow() -> sciline.Pipeline:
    """
    Create Sans2d tutorial workflow.

    Equivalent to :func:`Sans2dWorkflow`, but with loaders for tutorial data instead
    of Mantid-based loaders.
    """
    workflow = Sans2dWorkflow()
    workflow.insert(load_tutorial_run)
    workflow.insert(load_tutorial_direct_beam)
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import importlib.metadata

from . import workflow
from .larmor_workflow import (
    LokiAtLarmorTutorialWorkflow,
    LokiAtLarmorWorkflow,
    larmor_default_parameters,
)
from .workflow import LokiWorkflow, loki_default_parameters

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = [
    'LokiAtLarmorTutorialWorkflow',
    'LokiAtLarmorWorkflow',
    'LokiWorkflow',
    'larmor_default_parameters',
    'loki_default_parameters',
    'workflow',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import matplotlib
import pytest
import scipp as sc

import ess.loki.data  # noqa: F401
from ess import loki
from ess.loki.diagnostics import InstrumentView, LokiBankViewer
from ess.sans.types import (
    BeamCenter,
    Filename,
    NeXusDetectorName,
    RawDetector,
    SampleRun,
)


@pytest.fixture(scope='module')
def loki_data():
    wf = loki.LokiWorkflow()
    wf[BeamCenter] = sc.vector([0, 0, 0], unit='m')
    wf[Filename[SampleRun]] = loki.data.loki_coda_file_small()

    data = sc.DataGroup()
    for i in range(9):
        key = f"loki_detector_{i}"
        wf[NeXusDetectorName] = key
        data[key] = wf.compute(RawDetector[SampleRun])

    return data


def test_create_loki_bank_viewer(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist())
    assert len(viewer.tabs.children) == 9 + 1  # 9 banks + all banks tab


def test_loki_bank_viewer_plotting_args(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist(), norm='log', cmap='jet')
    for fig in viewer.subplots:
        mapper = fig.view.colormapper
        assert mapper.norm == 'log'
        assert mapper.cmap.name == 'jet'


def test_loki_bank_viewer_toggle_log_scale(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist())
    for fig in viewer.subplots:
        mapper = fig.view.colormapper
        assert mapper.norm == 'linear'
    viewer.log_button.value = True
    for fig in viewer.subplots:
        mapper = fig.view.colormapper
        assert mapper.norm == 'log'


def test_loki_bank_viewer_sum_all_layers(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist())
    old_max = [fig.view.colormapper.vmax for fig in viewer.subplots]
    viewer.layer_sum.value = True
    for i, fig in enumerate(viewer.subplots):
        assert fig.view.colormapper.vmax >= old_max[i]


def test_loki_bank_viewer_sum_all_straws(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist())
    old_max = [fig.view.colormapper.vmax for fig in viewer.subplots]
    viewer.straw_sum.value = True
    for i, fig in enumerate(viewer.subplots):
        assert fig.view.colormapper.vmax >= old_max[i]


def test_loki_bank_viewer_change_bank(loki_data):
    matplotlib.use('module://ipympl.backend_nbagg')
    viewer = LokiBankViewer(loki_data.hist())
    # For now, just check no error occurs when changing tab
    viewer.tabs.selected_index = 2
    # Change back to all banks
    viewer.tabs.selected_index = 0


def test_creat_loki_instrument_view(loki_data):
    InstrumentView(loki_data.hist())


def test_creat_loki_instrument_view_with_dim_slider(loki_data):
    InstrumentView(loki_data.hist(event_time_offset=200), dim='event_time_offset')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)


import pytest
import scipp as sc
from sciline import UnsatisfiedRequirement

from ess import loki
from ess.loki import LokiAtLarmorWorkflow
from ess.reduce import workflow
from ess.sans.types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BeamCenter,
    Filename,
    IntensityQ,
    LookupTableFilename,
    NeXusDetectorName,
    PixelMaskFilename,
    QBins,
    ReturnEvents,
    SampleRun,
    UncertaintyBroadcastMode,
    WavelengthDetector,
)


def test_sans_workflow_registers_subclasses():
    # Because it was imported
    assert LokiAtLarmorWorkflow in workflow.workflow_registry
    count = len(workflow.workflow_registry)

    @workflow.register_workflow
    class MyWorkflow: ...

    assert MyWorkflow in workflow.workflow_registry
    assert len(workflow.workflow_registry) == count + 1


def test_loki_larmor_workflow_parameters_returns_filtered_params():
    wf = LokiAtLarmorWorkflow()
    parameters = workflow.get_parameters(wf, (IntensityQ[SampleRun],))
    assert Filename[SampleRun] in parameters
    assert Filename[BackgroundRun] not in parameters


def test_loki_larmor_workflow_parameters_returns_no_params_for_no_outputs():
    wf = LokiAtLarmorWorkflow()
    parameters = workflow.get_parameters(wf, ())
    assert not parameters


def test_loki_larmor_workflow_parameters_with_param_returns_param():
    wf = LokiAtLarmorWorkflow()
    parameters = workflow.get_parameters(wf, (ReturnEvents,))
    assert parameters.keys() == {ReturnEvents}


def test_loki_larmor_workflow_compute_with_single_pixel_mask(larmor_workflow):
    wf = larmor_workflow(no_masks=False)
    wf[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    wf[PixelMaskFilename] = loki.data.loki_tutorial_mask_filenames()[0]
    # For simplicity, insert a fake beam center instead of computing it.
    wf[BeamCenter] = sc.vector([0.0, 0.0, 0.0], unit='m')

    result = wf.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)
    assert sc.identical(result.coords['Q'], wf.compute(QBins))
    assert result.sizes['Q'] == 100


@pytest.mark.parametrize("bank", list(range(9)))
def test_loki_workflow_needs_lookup_table(loki_workflow, bank):
    wf = loki_workflow()
    # For simplicity, insert a fake beam center instead of computing it.
    wf[BeamCenter] = sc.vector([0.0, 0.0, 0.0], unit='m')
    wf[NeXusDetectorName] = f'loki_detector_{bank}'
    with pytest.raises(UnsatisfiedRequirement, match='LookupTableFilename'):
        wf.compute(WavelengthDetector[SampleRun])


@pytest.mark.parametrize("bank", list(range(9)))
def test_loki_workflow_can_compute_wavelength(loki_workflow, bank):
    wf = loki_workflow()
    # For simplicity, insert a fake beam center instead of computing it.
    wf[BeamCenter] = sc.vector([0.0, 0.0, 0.0], unit='m')
    wf[NeXusDetectorName] = f'loki_detector_{bank}'
    wf[LookupTableFilename] = loki.data.loki_lookup_table_no_choppers()
    result = wf.compute(WavelengthDetector[SampleRun])
    assert 'wavelength' in result.bins.coords


@pytest.mark.parametrize("bank", list(range(9)))
def test_loki_workflow_can_compute_iofq(loki_workflow, bank):
    wf = loki_workflow()
    # For simplicity, insert a fake beam center instead of computing it.
    wf[BeamCenter] = sc.vector([0.0, 0.0, 0.0], unit='m')
    wf[NeXusDetectorName] = f'loki_detector_{bank}'
    wf[LookupTableFilename] = loki.data.loki_lookup_table_no_choppers()

    result = wf.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)
    assert sc.identical(result.coords['Q'], wf.compute(QBins))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest
import sciline
import scipp as sc

import ess.loki.data  # noqa: F401
from ess import loki
from ess.sans.types import (
    BackgroundRun,
    CorrectForGravity,
    DetectorMasks,
    DirectBeam,
    EmptyBeamRun,
    Filename,
    NeXusDetectorName,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    SampleRun,
    TransmissionRun,
    UncertaintyBroadcastMode,
    WavelengthBins,
)


@pytest.fixture
def larmor_workflow() -> sciline.Pipeline:
    def make_workflow(no_masks: bool = True) -> sciline.Pipeline:
        wf = loki.LokiAtLarmorWorkflow()

        wf[NeXusDetectorName] = 'larmor_detector'
        wf[Filename[SampleRun]] = loki.data.loki_tutorial_sample_run_60339()
        wf[Filename[BackgroundRun]] = loki.data.loki_tutorial_background_run_60393()
        wf[Filename[TransmissionRun[SampleRun]]] = (
            loki.data.loki_tutorial_sample_transmission_run()
        )
        wf[Filename[TransmissionRun[BackgroundRun]]] = (
            loki.data.loki_tutorial_run_60392()
        )
        wf[Filename[EmptyBeamRun]] = loki.data.loki_tutorial_run_60392()

        wf[WavelengthBins] = sc.linspace(
            'wavelength', start=1.0, stop=13.0, num=51, unit='angstrom'
        )
        wf[CorrectForGravity] = True
        wf[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
        wf[ReturnEvents] = False

        wf[QxBins] = sc.linspace('Qx', start=-0.3, stop=0.3, num=91, unit='1/angstrom')
        wf[QyBins] = sc.linspace('Qy', start=-0.2, stop=0.3, num=78, unit='1/angstrom')
        wf[QBins] = sc.linspace('Q', start=0.01, stop=0.3, num=101, unit='1/angstrom')
        # We have no direct-beam file for Loki currently
        wf[DirectBeam] = None
        if no_masks:
            wf[DetectorMasks] = {}

        return wf

    return make_workflow


@pytest.fixture
def loki_workflow() -> sciline.Pipeline:
    def make_workflow() -> sciline.Pipeline:
        wf = loki.LokiWorkflow()

        # For now, use a dummy file for all runs. This produces a result with all NaNs,
        # but is sufficient to test that the workflow runs end-to-end.
        # TODO: Replace with real test data when available.
        file = loki.data.loki_coda_file_one_event()
        wf[Filename[SampleRun]] = file
        wf[Filename[BackgroundRun]] = file
        wf[Filename[TransmissionRun[SampleRun]]] = file
        wf[Filename[TransmissionRun[BackgroundRun]]] = file
        wf[Filename[EmptyBeamRun]] = file

        wf[WavelengthBins] = sc.linspace(
            'wavelength', start=1.0, stop=13.0, num=51, unit='angstrom'
        )
        wf[CorrectForGravity] = True
        wf[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
        wf[ReturnEvents] = False

        wf[QxBins] = sc.linspace('Qx', start=-0.3, stop=0.3, num=91, unit='1/angstrom')
        wf[QyBins] = sc.linspace('Qy', start=-0.2, stop=0.3, num=78, unit='1/angstrom')
        wf[QBins] = sc.linspace('Q', start=0.01, stop=0.3, num=101, unit='1/angstrom')
        # We have no direct-beam file for Loki currently
        wf[DirectBeam] = None
        wf[DetectorMasks] = {}

        return wf

    return make_workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Providers for the ISIS instruments.
"""

from dataclasses import dataclass
from typing import Generic, NewType

import sciline
import scipp as sc
import scippneutron as scn
import scippnexus as snx

from ess.reduce.nexus.types import NeXusTransformation, Position
from ess.sans.types import (
    BeamCenter,
    DetectorIDs,
    DetectorPixelShape,
    DetectorPositionOffset,
    EmptyDetector,
    EmptyMonitor,
    Incident,
    Measurement,
    MonitorPositionOffset,
    MonitorType,
    NeXusComponent,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    RawDetector,
    RawMonitor,
    RunType,
    SampleRun,
    Transmission,
    WavelengthDetector,
    WavelengthMonitor,
)

from .io import LoadedFileContents
from .mantidio import Period


class MonitorOffset(sciline.Scope[MonitorType, sc.Variable], sc.Variable):
    """
    Offset for monitor position for all runs.
    """


@dataclass
class MonitorSpectrumNumber(Generic[MonitorType]):
    """
    Spectrum number for monitor data.

    This is used to identify the monitor in ISIS histogram data files.
    """

    value: int


DetectorBankOffset = NewType('DetectorBankOffset', sc.Variable)
SampleOffset = NewType('SampleOffset', sc.Variable)


def default_parameters() -> dict:
    return {
        MonitorOffset[Incident]: MonitorOffset[Incident](
            sc.vector([0, 0, 0], unit='m')
        ),
        MonitorOffset[Transmission]: MonitorOffset[Transmission](
            sc.vector([0, 0, 0], unit='m')
        ),
        DetectorBankOffset: DetectorBankOffset(sc.vector([0, 0, 0], unit='m')),
        SampleOffset: SampleOffset(sc.vector([0, 0, 0], unit='m')),
        NonBackgroundWavelengthRange: None,
        Period: None,
        # Not used for event files, setting default so the workflow works. If histogram
        # data is used, the user should set this to the correct value.
        MonitorSpectrumNumber[Incident]: MonitorSpectrumNumber[Incident](-1),
        MonitorSpectrumNumber[Transmission]: MonitorSpectrumNumber[Transmission](-1),
    }


def to_detector_position_offset(
    global_offset: DetectorBankOffset, beam_center: BeamCenter
) -> DetectorPositionOffset[RunType]:
    return DetectorPositionOffset[RunType](global_offset - beam_center)


def to_monitor_position_offset(
    global_offset: MonitorOffset[MonitorType],
) -> MonitorPositionOffset[RunType, MonitorType]:
    return MonitorPositionOffset[RunType, MonitorType](global_offset)


def get_source_position(
    dg: LoadedFileContents[RunType],
) -> Position[snx.NXsource, RunType]:
    """Get source position from raw data."""
    return Position[snx.NXsource, RunType](dg['data'].coords['source_position'])


def get_sample_position(
    dg: LoadedFileContents[RunType], offset: SampleOffset
) -> Position[snx.NXsample, RunType]:
    """Get sample position from raw data and apply user offset."""
    return Position[snx.NXsample, RunType](
        dg['data'].coords['sample_position'] + offset.to(unit='m')
    )


def get_detector_data(
    dg: LoadedFileContents[RunType],
) -> NeXusComponent[snx.NXdetector, RunType]:
    """Get detector data and apply user offsets to raw data.

    Parameters
    ----------
    dg:
        Data loaded with Mantid and converted to Scipp.
    """
    # The generic NeXus workflow will try to extract 'data' from this, which is exactly
    # what we also have in the Mantid data. We use the generic workflow since it also
    # applies offsets, etc.
    return NeXusComponent[snx.NXdetector, RunType](dg)


def get_calibrated_isis_detector(
    detector: NeXusComponent[snx.NXdetector, RunType],
    *,
    offset: DetectorPositionOffset[RunType],
) -> EmptyDetector[RunType]:
    """
    Replacement for :py:func:`ess.reduce.nexus.workflow.get_calibrated_detector`.

    Differences:

    - The detector position is already pre-computed.
    - The detector is not reshaped.

    The reason for the partial duplication is to avoid having to put ISIS/Mantid
    specific code in the generic workflow.
    """
    da = detector['data']
    position = detector['data'].coords['position']
    return EmptyDetector[RunType](
        da.assign_coords(position=position + offset.to(unit=position.unit))
    )


def get_monitor_data(
    dg: LoadedFileContents[RunType],
    nexus_name: NeXusMonitorName[MonitorType],
    spectrum_number: MonitorSpectrumNumber[MonitorType],
) -> NeXusComponent[MonitorType, RunType]:
    """
    Extract monitor data that was loaded together with detector data.

    If the raw file is histogram data, Mantid stores this as a Workspace2D, where some
    or all spectra correspond to monitors.

    Parameters
    ----------
    dg:
        Data loaded with Mantid and converted to Scipp.
    nexus_name:
        Name of the monitor in the NeXus file, e.g. 'incident_monitor' or
        'transmission_monitor'. Used when raw data is an EventWorkspace, where
        monitors are stored in a group with this name.
    spectrum_number:
        Spectrum number of the monitor in the NeXus file, e.g., 3 for incident monitor
        or 4 for transmission monitor. This is used when the raw data is a
        Workspace2D, where the monitor data is stored in a spectrum with this number.

    Returns
    -------
    :
        Monitor data extracted from the loaded file.
    """
    # The generic NeXus workflow will try to extract 'data' from this, which is exactly
    # what we also have in the Mantid data. We use the generic workflow since it also
    # applies offsets, etc.
    if 'monitors' in dg:
        # From EventWorkspace
        monitor = dg['monitors'][nexus_name]['data']
    else:
        # From Workspace2D
        monitor = sc.values(dg["data"]["spectrum", sc.index(spectrum_number.value)])
    return NeXusComponent[MonitorType, RunType](
        sc.DataGroup(data=monitor, position=monitor.coords['position'])
    )


def get_calibrated_isis_monitor(
    monitor: NeXusComponent[MonitorType, RunType],
    offset: MonitorPositionOffset[RunType, MonitorType],
) -> EmptyMonitor[RunType, MonitorType]:
    """
    Replacement for :py:func:`ess.reduce.nexus.workflow.get_calibrated_detector`.

    Differences:

    - The detector position is already pre-computed.
    - The detector is not reshaped.

    The reason for the partial duplication is to avoid having to put ISIS/Mantid
    specific code in the generic workflow.
    """
    da = monitor['data']
    position = monitor['data'].coords['position']
    return EmptyMonitor[RunType, MonitorType](
        da.assign_coords(position=position + offset.to(unit=position.unit))
    )


def dummy_assemble_detector_data(
    detector: EmptyDetector[RunType],
) -> RawDetector[RunType]:
    """Dummy assembly of detector data, detector already contains neutron data."""
    return RawDetector[RunType](detector)


def dummy_assemble_monitor_data(
    monitor: EmptyMonitor[RunType, MonitorType],
) -> RawMonitor[RunType, MonitorType]:
    """Dummy assembly of monitor data, monitor already contains neutron data."""
    return RawMonitor[RunType, MonitorType](monitor)


def _isis_convert_to_wavelength(da: sc.DataArray, scatter: bool) -> sc.DataArray:
    graph = {
        **scn.conversion.graph.beamline.beamline(scatter=scatter),
        **scn.conversion.graph.tof.elastic("tof"),
    }
    return da.transform_coords('wavelength', graph=graph, keep_intermediate=False)


def data_to_wavelength(da: RawDetector[RunType]) -> WavelengthDetector[RunType]:
    return WavelengthDetector[RunType](_isis_convert_to_wavelength(da, scatter=True))


def monitor_to_wavelength(
    da: RawMonitor[RunType, MonitorType],
) -> WavelengthMonitor[RunType, MonitorType]:
    return WavelengthMonitor[RunType, MonitorType](
        _isis_convert_to_wavelength(da, scatter=False)
    )


def experiment_metadata(dg: LoadedFileContents[RunType]) -> Measurement[RunType]:
    """Get experiment metadata from the raw sample data."""
    return Measurement[RunType](
        title=dg['run_title'].value,
        run_number=dg['run_number'],
    )


def helium3_tube_detector_pixel_shape() -> DetectorPixelShape[RunType]:
    # Pixel radius and length
    # found here:
    # https://github.com/mantidproject/mantid/blob/main/instrument/SANS2D_Definition_Tubes.xml
    R = 0.00405
    L = 0.002033984375
    pixel_shape = sc.DataGroup(
        {
            'vertices': sc.vectors(
                dims=['vertex'],
                values=[
                    # Coordinates in pixel-local coordinate system
                    # Bottom face center
                    [0, 0, 0],
                    # Bottom face edge
                    [R, 0, 0],
                    # Top face center
                    [0, L, 0],
                ],
                unit='m',
            ),
            'nexus_class': 'NXcylindrical_geometry',
        }
    )
    return pixel_shape


def lab_frame_transform() -> NeXusTransformation[snx.NXdetector, RunType]:
    # Rotate +y to -x
    return NeXusTransformation[snx.NXdetector, RunType](
        sc.spatial.rotation(value=[0, 0, 1 / 2**0.5, 1 / 2**0.5])
    )


def get_detector_ids_from_sample_run(data: RawDetector[SampleRun]) -> DetectorIDs:
    """Extract detector IDs from sample run.

    This overrides the function in the masking module which gets the detector IDs from
    the detector before loading event data. In this ISIS case files are loaded using
    Mantid which does not load event separately, so we get IDs from the data.
    """
    return DetectorIDs(
        data.coords[
            'detector_number' if 'detector_number' in data.coords else 'detector_id'
        ]
    )


providers = (
    dummy_assemble_detector_data,
    dummy_assemble_monitor_data,
    experiment_metadata,
    to_detector_position_offset,
    to_monitor_position_offset,
    get_source_position,
    get_sample_position,
    get_detector_data,
    get_calibrated_isis_detector,
    get_calibrated_isis_monitor,
    get_detector_ids_from_sample_run,
    get_monitor_data,
    data_to_wavelength,
    monitor_to_wavelength,
    lab_frame_transform,
    helium3_tube_detector_pixel_shape,
)



from argparse import ArgumentParser
from pathlib import Path

import tomli

parser = ArgumentParser()
parser.add_argument(
    "--nightly",
    default="",
    help="List of dependencies to install from main branch for nightly tests, "
    "separated by commas.",
)
args = parser.parse_args()

CUSTOM_AUTO_SEPARATOR = """
# --- END OF CUSTOM SECTION ---
# The following was generated by 'tox -e deps', DO NOT EDIT MANUALLY!
"""


def write_dependencies(dependency_name: str, dependencies: list[str]) -> None:
    path = Path(f"{dependency_name}.in")
    if path.exists():
        sections = path.read_text().split(CUSTOM_AUTO_SEPARATOR)
        if len(sections) > 1:
            custom = sections[0]
        else:
            custom = ""
    else:
        custom = ""
    with path.open("w") as f:
        f.write(custom)
        f.write(CUSTOM_AUTO_SEPARATOR)
        f.write("\n".join(dependencies))
        f.write("\n")


with open("../pyproject.toml", "rb") as toml_file:
    pyproject = tomli.load(toml_file)
    dependencies = pyproject["project"].get("dependencies")
    if dependencies is None:
        raise RuntimeError("No dependencies found in pyproject.toml")
    dependencies = [dep.strip().strip('"') for dep in dependencies]
    test_dependencies = (
        pyproject["project"].get("optional-dependencies", {}).get("test", [])
    )
    test_dependencies = [dep.strip().strip('"') for dep in test_dependencies]


write_dependencies("base", dependencies)
write_dependencies("basetest", test_dependencies)


def as_nightly(repo: str) -> str:
    if "/" in repo:
        org, repo = repo.split("/")
    else:
        org = "scipp"
    if repo == "scipp":
        # With the standard pip resolver index-url takes precedence over
        # extra-index-url but with uv it's reversed, so if we move to tox-uv
        # this needs to be reversed.
        return (
            "scipp\n"
            "--index-url=https://pypi.anaconda.org/scipp-nightly-wheels/simple/\n"
            "--extra-index-url=https://pypi.org/simple\n"
            "--pre"
        )
    return f"{repo} @ git+https://github.com/{org}/{repo}@main"


nightly = tuple(args.nightly.split(",") if args.nightly else [])
nightly_dependencies = [
    dep for dep in dependencies + test_dependencies if not dep.startswith(nightly)
]
nightly_dependencies += [as_nightly(arg) for arg in nightly]

write_dependencies("nightly", nightly_dependencies)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
This modules defines the domain types uses in esssans.

The domain types are used to define parameters and to request results from a Sciline
pipeline."""

from collections.abc import Sequence
from typing import NewType, TypeVar

import sciline
import scipp as sc

from ess.reduce.nexus import types as reduce_t
from ess.reduce.uncertainty import UncertaintyBroadcastMode as _UncertaintyBroadcastMode
from ess.reduce.unwrap import types as unwrap_t

BackgroundRun = reduce_t.BackgroundRun
EmptyDetector = reduce_t.EmptyDetector
EmptyMonitor = reduce_t.EmptyMonitor
RawDetector = reduce_t.RawDetector
DetectorPositionOffset = reduce_t.DetectorPositionOffset
EmptyBeamRun = reduce_t.EmptyBeamRun
Filename = reduce_t.Filename
GravityVector = reduce_t.GravityVector
Incident = reduce_t.IncidentMonitor
RawMonitor = reduce_t.RawMonitor
MonitorPositionOffset = reduce_t.MonitorPositionOffset
NeXusMonitorName = reduce_t.NeXusName
NeXusComponent = reduce_t.NeXusComponent
NeXusTransformation = reduce_t.NeXusTransformation
Position = reduce_t.Position
SampleRun = reduce_t.SampleRun
Transmission = reduce_t.TransmissionMonitor
TransmissionRun = reduce_t.TransmissionRun

LookupTableRelativeErrorThreshold = unwrap_t.LookupTableRelativeErrorThreshold
LookupTableFilename = unwrap_t.LookupTableFilename
LookupTable = unwrap_t.LookupTable
WavelengthMonitor = unwrap_t.WavelengthMonitor
WavelengthDetector = unwrap_t.WavelengthDetector

DetectorBankSizes = reduce_t.DetectorBankSizes
NeXusDetectorName = reduce_t.NeXusDetectorName

MonitorType = TypeVar('MonitorType', Incident, Transmission)
RunType = reduce_t.RunType


UncertaintyBroadcastMode = _UncertaintyBroadcastMode

# 1.3  Numerator and denominator of IofQ
Numerator = NewType('Numerator', sc.DataArray)
"""Numerator of IntensityQ"""
Denominator = NewType('Denominator', sc.DataArray)
"""Denominator of IntensityQ"""
IofQPart = TypeVar('IofQPart', Numerator, Denominator)
"""TypeVar used for specifying Numerator or Denominator of IntensityQ"""

# 1.4  Entry paths in NeXus files
PixelShapePath = NewType('PixelShapePath', str)
"""
Name of the entry where the pixel shape is stored in the NeXus file
"""

TransformationPath = NewType('TransformationPath', str)
"""
Name of the entry where the transformation computed from the
transformation chain is stored, for the detectors and the monitors
"""

# 2  Workflow parameters

ReturnEvents = NewType('ReturnEvents', bool)
"""Whether to return events in the output I(Q)"""

WavelengthBins = NewType('WavelengthBins', sc.Variable)
"""Wavelength binning"""

WavelengthBands = NewType('WavelengthBands', sc.Variable | None)
"""Wavelength bands. Typically a single band, set to first and last value of
WavelengthBins.

The wavelength bands can however be used to compute the scattering cross section inside
multiple wavelength bands. In this case, the wavelength bands must be either one- or
two-dimensional.

In the case of a one-dimensional array, the values represent a set of non-overlapping
bins in the wavelength dimension.

In the case of a two-dimensional array, the values represent a set of (possibly
overlapping or non-contiguous) wavelength bands, with the first dimension being the
band index and the second dimension being the wavelength. For each band, there must be
two wavelength values defining the start and end wavelength of the band.
"""

ProcessedWavelengthBands = NewType('ProcessedWavelengthBands', sc.Variable)
"""Processed wavelength bands, as a two-dimensional variable, with the first dimension
being the band index and the second dimension being the wavelength. For each band, there
must be two wavelength values defining the start and end wavelength of the band."""


QBins = NewType('QBins', sc.Variable)
"""Q binning used when computing IofQ"""

QxBins = NewType('QxBins', sc.Variable)
"""Qx binning used when computing IofQxy"""

QyBins = NewType('QyBins', sc.Variable)
"""Qy binning used when computing IofQxy"""

NonBackgroundWavelengthRange = NewType(
    'NonBackgroundWavelengthRange', sc.Variable | None
)
"""Range of wavelengths that are not considered background in the monitor"""

DirectBeamFilename = NewType('DirectBeamFilename', str)
"""Filename of direct beam correction"""

BeamCenter = NewType('BeamCenter', sc.Variable)
"""Beam center, may be set directly or computed using beam-center finder"""

WavelengthMask = NewType('WavelengthMask', sc.DataArray | None)
"""Optional wavelength mask"""

CorrectForGravity = NewType('CorrectForGravity', bool)
"""Whether to correct for gravity when computing wavelength and Q."""

DimsToKeep = NewType('DimsToKeep', Sequence[str])
"""Dimensions that should not be reduced and thus still be present in the final
I(Q) result (this is typically the layer dimension)."""

OutFilename = NewType('OutFilename', str)
"""Filename of the output"""


PixelMaskFilename = NewType('PixelMaskFilename', str)

DetectorIDs = NewType('DetectorIDs', sc.Variable)
"""1-D variable listing all detector IDs."""

MaskedDetectorIDs = NewType('MaskedDetectorIDs', sc.Variable)
"""1-D variable listing all masked detector IDs."""


DetectorMasks = NewType('DetectorMasks', dict[str, sc.Variable])
"""Masks for detector pixels"""


# 3  Workflow (intermediate) results


DirectBeam = NewType('DirectBeam', sc.DataArray | None)
"""Direct beam"""


class TransmissionFraction(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Transmission fraction"""


CleanDirectBeam = NewType('CleanDirectBeam', sc.DataArray)
"""Direct beam after resampling to required wavelength bins, else and array of ones."""


class DetectorPixelShape(sciline.Scope[RunType, sc.DataGroup], sc.DataGroup):
    """Geometry of the detector from description in nexus file."""


class SolidAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Solid angle of detector pixels seen from sample position"""


PixelMask = NewType('PixelMask', sc.Variable)


class MonitorTerm(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Monitor-dependent factor of the Normalization term (numerator) for IofQ."""


class CorrectedDetector(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """
    Data with masks and corrections applied, used for numerator or denominator of IofQ.
    """


class DetectorTerm(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """
    Prerequisite for IofQ numerator or denominator.

    This can either be the sample or background counts, converted to wavelength,
    or the respective normalization terms computed from the respective solid angle,
    direct beam, and monitors.
    """


class NormalizedQ(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of applying wavelength scaling/masking to :py:class:`BinnedQ`"""


class NormalizedQxQy(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of applying wavelength scaling/masking to :py:class:`BinnedQxQy`"""


class QDetector(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of converting :py:class:`DetectorTermMasked` to Q"""


class QxyDetector(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of converting :py:class:`DetectorTermMasked` to Qx and Qy"""


class BinnedQ(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of histogramming/binning :py:class:`QDetector` over all pixels into Q
    bins"""


class BinnedQxQy(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of histogramming/binning :py:class:`QxyDetector` over all pixels into Qx
    and Qy bins"""


class ReducedQ(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of reducing :py:class:`BinnedQ` over the wavelength dimensions"""


class ReducedQxQy(sciline.Scope[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of reducing :py:class:`BinnedQxQy` over the wavelength dimensions"""


class IntensityQ(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """I(Q)"""


class IntensityQxQy(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """I(Qx, Qy)"""


BackgroundSubtractedIofQ = NewType('BackgroundSubtractedIofQ', sc.DataArray)
"""I(Q) with background (given by I(Q) of the background run) subtracted"""

BackgroundSubtractedIofQxy = NewType('BackgroundSubtractedIofQxy', sc.DataArray)
"""I(Qx, Qy) with background (given by I(Qx, Qy) of the background run) subtracted"""


class CorrectedMonitor(sciline.Scope[RunType, MonitorType, sc.DataArray], sc.DataArray):
    """Monitor data cleaned of background counts"""


# 4  Metadata

Beamline = reduce_t.Beamline
Measurement = reduce_t.Measurement



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from collections.abc import Hashable, Iterable

import pandas as pd
import sciline
import scipp as sc

from ess.reduce.parameter import parameter_mappers
from ess.reduce.unwrap import GenericUnwrapWorkflow

from . import common, conversions, i_of_q, masking, normalization
from .types import (
    BackgroundRun,
    CorrectForGravity,
    Denominator,
    DetectorBankSizes,
    DetectorMasks,
    DimsToKeep,
    EmptyBeamRun,
    Filename,
    Incident,
    NeXusDetectorName,
    NormalizedQ,
    NormalizedQxQy,
    Numerator,
    PixelMaskFilename,
    SampleRun,
    TransformationPath,
    Transmission,
    TransmissionRun,
    WavelengthBands,
    WavelengthMask,
)


def _merge(*dicts: dict) -> dict:
    return {key: value for d in dicts for key, value in d.items()}


def merge_contributions(*data: sc.DataArray) -> sc.DataArray:
    if len(data) == 1:
        return data[0]
    reducer = sc.reduce(data)
    return reducer.bins.concat() if data[0].bins is not None else reducer.sum()


def with_pixel_mask_filenames(
    workflow: sciline.Pipeline, masks: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with pixel mask filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    masks:
        List or tuple of pixel mask filenames to set.
    """
    workflow = workflow.copy()
    workflow[DetectorMasks] = (
        workflow[DetectorMasks]
        .map(pd.DataFrame({PixelMaskFilename: masks}).rename_axis('mask'))
        .reduce(index='mask', func=_merge)
    )
    return workflow


def with_banks(
    workflow: sciline.Pipeline,
    banks: Iterable[str],
    index: Iterable[Hashable] | None = None,
) -> sciline.Pipeline:
    """
    Return modified workflow with bank names set.

    Since banks typically have different Q-resolution the I(Q) of banks are not merged.
    That is, the resulting workflow will have separate outputs for each bank. Use
    :py:func:`sciline.compute_mapped` to compute results for all banks.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    banks:
        List or tuple of bank names to set.
    index:
        Index to use for the DataFrame. If not provided, the bank names are used.
    """
    index = index or banks
    return workflow.map(
        pd.DataFrame({NeXusDetectorName: banks}, index=index).rename_axis('bank')
    )


def _set_runs(
    pipeline: sciline.Pipeline, runs: Iterable[str], key: Hashable, axis_name: str
) -> sciline.Pipeline:
    pipeline = pipeline.copy()
    runs = pd.DataFrame({Filename[key]: runs}).rename_axis(axis_name)
    for part in (Numerator, Denominator):
        for qtype in (NormalizedQ, NormalizedQxQy):
            pipeline[qtype[key, part]] = (
                pipeline[qtype[key, part]]
                .map(runs)
                .reduce(index=axis_name, func=merge_contributions)
            )
    return pipeline


def with_sample_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with sample run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of sample run filenames to set.
    """
    return _set_runs(workflow, runs, SampleRun, 'sample_run')


def with_background_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with background run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of background run filenames to set.
    """
    return _set_runs(workflow, runs, BackgroundRun, 'background_run')


parameter_mappers[PixelMaskFilename] = with_pixel_mask_filenames
# TODO: for now, we leave the mapping over detector banks out, because we do not have a
# method to merge the I(Q) of different banks, and we thus cannot compute a single
# result from the workflow. So only a single detector bank can be processed at a time.
# parameter_mappers[NeXusDetectorName] = with_banks
parameter_mappers[Filename[SampleRun]] = with_sample_runs
parameter_mappers[Filename[BackgroundRun]] = with_background_runs


providers = (
    *conversions.providers,
    *i_of_q.providers,
    *masking.providers,
    *normalization.providers,
    common.beam_center_to_detector_position_offset,
)
"""
List of providers for setting up a Sciline pipeline.

This provides a default workflow, including a beam-center estimation based on a
center-of-mass approach. Providers for loadings files are not included. Combine with
the providers for a specific instrument, such as :py:data:`esssans.sans2d.providers`
to setup a complete workflow.
"""


def SansWorkflow() -> sciline.Pipeline:
    """
    Common base for SANS workflows.

    Returns
    -------
    :
        SANS workflow as a sciline.Pipeline
    """
    workflow = GenericUnwrapWorkflow(
        run_types=(
            SampleRun,
            EmptyBeamRun,
            BackgroundRun,
            TransmissionRun[SampleRun],
            TransmissionRun[BackgroundRun],
        ),
        monitor_types=(Incident, Transmission),
    )
    for provider in providers:
        workflow.insert(provider)
    workflow[CorrectForGravity] = CorrectForGravity(False)
    workflow[DetectorBankSizes] = DetectorBankSizes({})
    workflow[DimsToKeep] = DimsToKeep(())
    workflow[TransformationPath] = TransformationPath('transform')
    workflow[WavelengthBands] = WavelengthBands(None)
    workflow[WavelengthMask] = WavelengthMask(None)
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from ess import isissans, loki, sans

"""Tests of package integrity.

Note that additional imports need to be added for repositories that
contain multiple packages.
"""


def test_has_version():
    assert hasattr(isissans, '__version__')
    assert hasattr(loki, '__version__')
    assert hasattr(sans, '__version__')


if __name__ == '__main__':
    test_has_version()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import scipp as sc
from scipp.scipy.interpolate import interp1d

from ess.reduce.uncertainty import UncertaintyBroadcastMode, broadcast_uncertainties

from .common import mask_range
from .logging import get_logger
from .types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BackgroundSubtractedIofQxy,
    BinnedQ,
    BinnedQxQy,
    CleanDirectBeam,
    CorrectedMonitor,
    DimsToKeep,
    DirectBeam,
    IntensityQ,
    IntensityQxQy,
    IofQPart,
    MonitorType,
    NonBackgroundWavelengthRange,
    QBins,
    QDetector,
    QxBins,
    QxyDetector,
    QyBins,
    ReturnEvents,
    RunType,
    SampleRun,
    WavelengthBins,
    WavelengthMonitor,
)


def preprocess_monitor_data(
    monitor: WavelengthMonitor[RunType, MonitorType],
    wavelength_bins: WavelengthBins,
    non_background_range: NonBackgroundWavelengthRange,
    uncertainties: UncertaintyBroadcastMode,
) -> CorrectedMonitor[RunType, MonitorType]:
    """
    Prepare monitor data for computing the transmission fraction.
    The input data are first converted to wavelength (if needed).
    If a ``non_background_range`` is provided, it defines the region where data is
    considered not to be background, and regions outside are background. A mean
    background level will be computed from the background and will be subtracted from
    the non-background counts.
    Finally, if wavelength bins are provided, the data is rebinned to match the
    requested binning.

    Parameters
    ----------
    monitor:
        The monitor to be pre-processed.
    wavelength_bins:
        The binning in wavelength to use for the rebinning.
    non_background_range:
        The range of wavelengths that defines the data which does not constitute
        background. Everything outside this range is treated as background counts.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`ess.reduce.uncertainty.UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        The input monitors converted to wavelength, cleaned of background counts, and
        rebinned to the requested wavelength binning.
    """
    background = None
    if non_background_range is not None:
        mask = sc.DataArray(
            data=sc.array(dims=[non_background_range.dim], values=[True]),
            coords={non_background_range.dim: non_background_range},
        )
        background = mask_range(monitor, mask=mask).mean()

    if monitor.bins is not None:
        monitor = monitor.hist(wavelength=wavelength_bins)
    else:
        monitor = monitor.rebin(wavelength=wavelength_bins)

    if background is not None:
        monitor -= broadcast_uncertainties(
            background, prototype=monitor, mode=uncertainties
        )

    # If the position-related coords are present, we need to remove them to allow for
    # dividing counts in different monitors to compute the transmission fraction.
    monitor = monitor.drop_coords({'position', 'Ltotal'} & set(monitor.coords))

    return CorrectedMonitor(monitor)


def resample_direct_beam(
    direct_beam: DirectBeam, wavelength_bins: WavelengthBins
) -> CleanDirectBeam:
    """
    If the wavelength binning of the direct beam function does not match the requested
    ``wavelength_bins``, perform a 1d interpolation of the function onto the bins.

    Parameters
    ----------
    direct_beam:
        The DataArray containing the direct beam function (it should have a dimension
        of wavelength).
    wavelength_bins:
        The binning in wavelength that the direct beam function should be resampled to.

    Returns
    -------
    :
        The direct beam function resampled to the requested resolution.
    """
    if direct_beam is None:
        return CleanDirectBeam(
            sc.DataArray(
                sc.ones(dims=wavelength_bins.dims, shape=[len(wavelength_bins) - 1]),
                coords={'wavelength': wavelength_bins},
            )
        )
    if sc.identical(direct_beam.coords['wavelength'], wavelength_bins):
        return direct_beam
    if direct_beam.variances is not None:
        logger = get_logger('sans')
        logger.warning(
            'An interpolation is being performed on the direct_beam function. '
            'The variances in the direct_beam function will be dropped.'
        )
    func = interp1d(
        sc.values(direct_beam),
        'wavelength',
        fill_value="extrapolate",
        bounds_error=False,
    )
    return CleanDirectBeam(func(wavelength_bins, midpoints=True))


def bin_in_q(
    data: QDetector[RunType, IofQPart],
    q_bins: QBins,
    dims_to_keep: DimsToKeep,
) -> BinnedQ[RunType, IofQPart]:
    """
    Merges data from all pixels into a single I(Q) spectrum:

    * In the case of event data, events in all bins are concatenated
    * In the case of dense data, counts in all spectra are summed

    Parameters
    ----------
    data:
        A DataArray containing the data that is to be converted to Q.
    q_bins:
        The binning in Q to be used.
    dims_to_keep:
        Dimensions that should not be reduced and thus still be present in the final
        I(Q) result (this is typically the layer dimension).

    Returns
    -------
    :
        The input data converted to Q and then summed over all detector pixels.
    """
    out = _bin_in_q(data=data, edges={'Q': q_bins}, dims_to_keep=dims_to_keep)
    return BinnedQ[RunType, IofQPart](out)


def bin_in_qxy(
    data: QxyDetector[RunType, IofQPart],
    qx_bins: QxBins,
    qy_bins: QyBins,
    dims_to_keep: DimsToKeep,
) -> BinnedQxQy[RunType, IofQPart]:
    """
    Merges data from all pixels into a single I(Q) spectrum:

    * In the case of event data, events in all bins are concatenated
    * In the case of dense data, counts in all spectra are summed

    Parameters
    ----------
    data:
        A DataArray containing the data that is to be converted to Q.
    qx_bins:
        The binning in Qx to be used.
    qy_bins:
        The binning in Qy to be used.
    dims_to_keep:
        Dimensions that should not be reduced and thus still be present in the final
        I(Q) result (this is typically the layer dimension).

    Returns
    -------
    :
        The input data converted to Qx and Qy and then summed over all detector pixels.
    """
    # We make Qx the inner dim, such that plots naturally show Qx on the x-axis.
    out = _bin_in_q(
        data=data,
        edges={'Qy': qy_bins, 'Qx': qx_bins},
        dims_to_keep=dims_to_keep,
    )
    return BinnedQxQy[RunType, IofQPart](out)


def _bin_in_q(
    data: sc.DataArray, edges: dict[str, sc.Variable], dims_to_keep: tuple[str, ...]
) -> sc.DataArray:
    dims_to_reduce = set(data.dims) - {'wavelength'} - set(dims_to_keep or ())
    return (data.hist if data.bins is None else data.bin)(**edges, dim=dims_to_reduce)


def _subtract_background(
    sample: sc.DataArray,
    background: sc.DataArray,
    return_events: ReturnEvents,
) -> sc.DataArray:
    if return_events and sample.bins is not None and background.bins is not None:
        return sample.bins.concatenate(-background)
    if sample.bins is not None:
        sample = sample.bins.sum()
    if background.bins is not None:
        background = background.bins.sum()
    return sample - background


def subtract_background(
    sample: IntensityQ[SampleRun],
    background: IntensityQ[BackgroundRun],
    return_events: ReturnEvents,
) -> BackgroundSubtractedIofQ:
    return BackgroundSubtractedIofQ(
        _subtract_background(
            sample=sample, background=background, return_events=return_events
        )
    )


def subtract_background_xy(
    sample: IntensityQxQy[SampleRun],
    background: IntensityQxQy[BackgroundRun],
    return_events: ReturnEvents,
) -> BackgroundSubtractedIofQxy:
    return BackgroundSubtractedIofQxy(
        _subtract_background(
            sample=sample, background=background, return_events=return_events
        )
    )


providers = (
    preprocess_monitor_data,
    resample_direct_beam,
    bin_in_q,
    bin_in_qxy,
    subtract_background,
    subtract_background_xy,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import numpy as np
import scipp as sc
from sciline import Pipeline

from .i_of_q import resample_direct_beam
from .types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    Denominator,
    DirectBeam,
    NormalizedQ,
    Numerator,
    ProcessedWavelengthBands,
    SampleRun,
    WavelengthBands,
    WavelengthBins,
)


def _compute_efficiency_correction(
    iofq_full: sc.DataArray,
    iofq_bands: sc.DataArray,
    wavelength_band_dim: str,
    I0: sc.Variable,
) -> sc.DataArray:
    """
    Compute the factor by which to multiply the direct beam function inside each
    wavelength band so that the $I(Q)$ curves for the full wavelength range and inside
    the bands overlap.

    Parameters
    ----------
    iofq_full:
        The $I(Q)$ for the full wavelength range.
    iofq_bands:
        The $I(Q)$ for the wavelength bands.
    wavelength_band_dim:
        The name of the wavelength band dimension.
    I0:
        The intensity of the I(Q) for the known sample at the lowest Q value.
    """
    invalid = (iofq_bands.data <= sc.scalar(0.0)) | ~sc.isfinite(iofq_bands.data)
    data = np.where(invalid.values, np.nan, (iofq_bands.data / iofq_full.data).values)
    eff = np.nanmedian(data, axis=iofq_bands.dims.index('Q'))

    scaling = sc.values(iofq_full['Q', 0].data) / I0
    # Note: do not use a `set` here because the order of dimensions is important
    dims = [dim for dim in iofq_bands.dims if dim != 'Q']
    out = sc.array(dims=dims, values=eff) * scaling
    return out.rename_dims({wavelength_band_dim: 'wavelength'})


def direct_beam(*, workflow: Pipeline, I0: sc.Variable, niter: int = 5) -> list[dict]:
    """
    Compute the direct beam function.

    Procedure:

    The idea behind the direct beam iterations is to determine an efficiency of the
    detectors as a function of wavelength.
    To calculate this, it is possible to compute $I(Q)$ for the full wavelength range,
    and for individual slices (bands) of the wavelength range.
    If the direct beam function used in the $I(Q)$ computation is correct, then $I(Q)$
    curves for the full wavelength range and inside the bands should overlap.

    We require two pipelines, one for the full wavelength range and one for the bands.

    The steps are as follows:

     1. Create a flat direct beam function, as a function of wavelength, with
        wavelength bins corresponding to the wavelength bands
     2. Calculate inside each band by how much one would have to multiply the final
        $I(Q)$ so that the curve would overlap with the full range curve
     3. Multiply the direct beam values inside each wavelength band by this factor
     4. Compare the full-range $I(Q)$ to a theoretical reference and add the
        corresponding additional scaling to the direct beam function
     5. Iterate a given number of times (typically less than 10) so as to gradually
        converge on a direct beam function

    TODO: For simplicity and robustness, we currently specify the number of times to
    iterate. We could imagine in the future having a convergence criterion instead to
    determine when to stop iterating.

    Parameters
    ----------
    pipeline:
        The pipeline to compute the differential scattering cross section I(Q).
    I0:
        The intensity of the I(Q) for the known sample at the lowest Q value.
    niter:
        The number of iterations to perform.
    """

    direct_beam_function = None
    bands = workflow.compute(ProcessedWavelengthBands)
    band_dim = (set(bands.dims) - {'wavelength'}).pop()
    full_wavelength_range = sc.concat([bands.min(), bands.max()], dim='wavelength')

    workflow = workflow.copy()
    workflow[DirectBeam] = None

    wavelength_bins = workflow.compute(WavelengthBins)
    parts = (
        NormalizedQ[SampleRun, Numerator],
        NormalizedQ[SampleRun, Denominator],
        NormalizedQ[BackgroundRun, Numerator],
        NormalizedQ[BackgroundRun, Denominator],
    )
    parts = workflow.compute(parts)
    # Convert events to histograms to make normalization (in every iteration) cheap
    for key in [
        NormalizedQ[SampleRun, Numerator],
        NormalizedQ[BackgroundRun, Numerator],
    ]:
        parts[key] = parts[key].hist(wavelength=wavelength_bins)

    # For now we simply strip all uncertainties, since direct beam function is
    # computed without variances anyway.
    parts = {key: sc.values(result) for key, result in parts.items()}
    for key, part in parts.items():
        workflow[key] = part
    sample0 = parts[NormalizedQ[SampleRun, Denominator]]
    background0 = parts[NormalizedQ[BackgroundRun, Denominator]]

    results = []

    for _it in range(niter):
        # The first time we compute I(Q), the direct beam function is not in the
        # parameters, nor given by any providers, so it will be considered flat.
        workflow[WavelengthBands] = full_wavelength_range
        iofq_full = workflow.compute(BackgroundSubtractedIofQ)
        workflow[WavelengthBands] = bands
        iofq_bands = workflow.compute(BackgroundSubtractedIofQ)

        if direct_beam_function is None:
            # Make a flat direct beam
            dims = [dim for dim in iofq_bands.dims if dim != 'Q']
            direct_beam_function = sc.DataArray(
                data=sc.ones(sizes={dim: iofq_bands.sizes[dim] for dim in dims}),
                coords={band_dim: sc.midpoints(bands, dim='wavelength').squeeze()},
            ).rename({band_dim: 'wavelength'})

        direct_beam_function *= _compute_efficiency_correction(
            iofq_full=iofq_full,
            iofq_bands=iofq_bands,
            wavelength_band_dim=band_dim,
            I0=I0,
        )

        # Scale denominator terms that were initially computed without direct beam
        # with the current direct beam function.
        db = resample_direct_beam(
            direct_beam=direct_beam_function,
            wavelength_bins=wavelength_bins,
        )
        db.coords['wavelength'] = sc.midpoints(
            db.coords['wavelength'], dim='wavelength'
        )
        workflow[NormalizedQ[SampleRun, Denominator]] = sample0 * db
        workflow[NormalizedQ[BackgroundRun, Denominator]] = background0 * db

        results.append(
            {
                'iofq_full': iofq_full,
                'iofq_bands': iofq_bands,
                'direct_beam': direct_beam_function,
            }
        )
    return results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import sciline
import scipp as sc
import scippnexus as snx
from scippneutron.conversion.beamline import (
    beam_aligned_unit_vectors,
    scattering_angles_with_gravity,
)
from scippneutron.conversion.graph import beamline, tof

from ess.reduce.uncertainty import broadcast_uncertainties

from .common import mask_range
from .types import (
    BinnedQ,
    BinnedQxQy,
    CorrectForGravity,
    Denominator,
    DetectorTerm,
    GravityVector,
    IofQPart,
    MonitorTerm,
    NormalizedQ,
    NormalizedQxQy,
    Numerator,
    Position,
    QDetector,
    QxyDetector,
    RunType,
    UncertaintyBroadcastMode,
    WavelengthMask,
)


def cyl_unit_vectors(incident_beam: sc.Variable, gravity: sc.Variable):
    vectors = beam_aligned_unit_vectors(incident_beam=incident_beam, gravity=gravity)
    return {
        'cyl_x_unit_vector': vectors['beam_aligned_unit_x'],
        'cyl_y_unit_vector': vectors['beam_aligned_unit_y'],
    }


def cylindrical_x(
    cyl_x_unit_vector: sc.Variable, scattered_beam: sc.Variable
) -> sc.Variable:
    """
    Compute the horizontal x coordinate perpendicular to the incident beam direction.
    Note that it is assumed here that the incident beam is perpendicular to the gravity
    vector.
    """
    return sc.dot(scattered_beam, cyl_x_unit_vector)


def cylindrical_y(
    cyl_y_unit_vector: sc.Variable, scattered_beam: sc.Variable
) -> sc.Variable:
    """
    Compute the vertical y coordinate perpendicular to the incident beam direction.
    Note that it is assumed here that the incident beam is perpendicular to the gravity
    vector.
    """
    return sc.dot(scattered_beam, cyl_y_unit_vector)


def phi_no_gravity(
    cylindrical_x: sc.Variable, cylindrical_y: sc.Variable
) -> sc.Variable:
    """
    Compute the cylindrical phi angle around the incident beam. Note that it is assumed
    here that the incident beam is perpendicular to the gravity vector.
    """
    return sc.atan2(y=cylindrical_y, x=cylindrical_x)


def Qxy(Q: sc.Variable, phi: sc.Variable) -> dict[str, sc.Variable]:
    """
    Compute the Qx and Qy components of the scattering vector from the scattering angle,
    wavelength, and phi angle.
    """
    Qx = sc.cos(phi)
    Qy = sc.sin(phi)
    if Q.bins is not None and phi.bins is not None:
        Qx *= Q
        Qy *= Q
    else:
        Qx = Qx * Q
        Qy = Qy * Q
    return {'Qx': Qx, 'Qy': Qy}


class ElasticCoordTransformGraph(sciline.Scope[RunType, dict], dict):
    """
    Coordinate transformation graph for SANS elastic scattering (which possibly
    includes the effects of gravitationaly pull on the neutrons).
    See :func:`sans_elastic` for more details.
    """


def sans_elastic(
    correct_for_gravity: CorrectForGravity,
    *,
    sample_position: Position[snx.NXsample, RunType],
    source_position: Position[snx.NXsource, RunType],
    gravity: GravityVector,
) -> ElasticCoordTransformGraph[RunType]:
    """
    Generate a coordinate transformation graph for SANS elastic scattering.

    It is based on classical conversions from ``tof`` and pixel ``position`` to
    :math:`\\lambda` (``wavelength``), :math:`\\theta` (``theta``) and
    :math:`Q` (``Q``), but can take into account the Earth's gravitational field,
    which bends the flight path of the neutrons, to compute the scattering angle
    :math:`\\theta`.

    The angle can be found using the following expression
    (`Seeger & Hjelm 1991 <https://doi.org/10.1107/S0021889891004764>`_):

    .. math::

       \\theta = \\frac{1}{2}\\sin^{-1}\\left(\\frac{\\sqrt{ x^{2} + \\left( y + \\frac{g m_{\\rm n}}{2 h^{2}} \\lambda^{2} L_{2}^{2} \\right)^{2} } }{L_{2}}\\right)

    where :math:`x` and :math:`y` are the spatial coordinates of the pixels in the
    horizontal and vertical directions, respectively,
    :math:`m_{\\rm n}` is the neutron mass,
    :math:`L_{2}` is the distance between the sample and a detector pixel,
    :math:`g` is the acceleration due to gravity,
    and :math:`h` is Planck's constant.

    By default, the effects of gravity on the neutron flight paths are not included
    (equivalent to :math:`g = 0` in the expression above).

    Parameters
    ----------
    correct_for_gravity:
        Take into account the bending of the neutron flight paths from the
        Earth's gravitational field if ``True``.
    gravity:
        A vector indicating the strength and direction of gravity.
        Required even if ``correct_for_gravity`` is ``False``.
    sample_position:
        Position of the sample as a vector.
    source_position:
        Position of the source as a vector.
    """  # noqa: E501
    graph = {
        **beamline.beamline(scatter=True),
        **tof.elastic_Q('wavelength'),
        'sample_position': lambda: sample_position,
        'source_position': lambda: source_position,
        'gravity': lambda: gravity,
    }
    if correct_for_gravity:
        del graph['two_theta']
        graph[('two_theta', 'phi')] = scattering_angles_with_gravity
    else:
        graph['phi'] = phi_no_gravity
    graph[('cyl_x_unit_vector', 'cyl_y_unit_vector')] = cyl_unit_vectors
    graph['cylindrical_x'] = cylindrical_x
    graph['cylindrical_y'] = cylindrical_y
    graph[('Qx', 'Qy')] = Qxy
    return ElasticCoordTransformGraph[RunType](graph)


def mask_wavelength_q(
    da: BinnedQ[RunType, Numerator], mask: WavelengthMask
) -> NormalizedQ[RunType, Numerator]:
    if mask is not None:
        da = mask_range(da, mask=mask)
    return NormalizedQ[RunType, Numerator](da)


def mask_wavelength_qxy(
    da: BinnedQxQy[RunType, Numerator], mask: WavelengthMask
) -> NormalizedQxQy[RunType, Numerator]:
    if mask is not None:
        da = mask_range(da, mask=mask)
    return NormalizedQxQy[RunType, Numerator](da)


def mask_and_scale_wavelength_q(
    da: BinnedQ[RunType, Denominator],
    mask: WavelengthMask,
    wavelength_term: MonitorTerm[RunType],
    uncertainties: UncertaintyBroadcastMode,
) -> NormalizedQ[RunType, Denominator]:
    da = da * broadcast_uncertainties(wavelength_term, prototype=da, mode=uncertainties)
    if mask is not None:
        da = mask_range(da, mask=mask)
    return NormalizedQ[RunType, Denominator](da)


def mask_and_scale_wavelength_qxy(
    da: BinnedQxQy[RunType, Denominator],
    mask: WavelengthMask,
    wavelength_term: MonitorTerm[RunType],
    uncertainties: UncertaintyBroadcastMode,
) -> NormalizedQxQy[RunType, Denominator]:
    da = da * broadcast_uncertainties(wavelength_term, prototype=da, mode=uncertainties)
    if mask is not None:
        da = mask_range(da, mask=mask)
    return NormalizedQxQy[RunType, Denominator](da)


def _compute_Q(
    data: sc.DataArray, graph: ElasticCoordTransformGraph, target: tuple[str, ...]
) -> sc.DataArray:
    # Keep naming of wavelength dim, subsequent steps use a (Q[xy], wavelength) binning.
    return QDetector[RunType, IofQPart](
        data.transform_coords(
            target,
            graph=graph,
            keep_intermediate=False,
            rename_dims=False,
        )
    )


def compute_Q(
    data: DetectorTerm[RunType, IofQPart],
    graph: ElasticCoordTransformGraph[RunType],
) -> QDetector[RunType, IofQPart]:
    """
    Convert a data array from wavelength to Q.
    """
    return QDetector[RunType, IofQPart](
        _compute_Q(data=data, graph=graph, target=('Q',))
    )


def compute_Qxy(
    data: DetectorTerm[RunType, IofQPart],
    graph: ElasticCoordTransformGraph[RunType],
) -> QxyDetector[RunType, IofQPart]:
    """
    Convert a data array from wavelength to Qx and Qy.
    """
    return QxyDetector[RunType, IofQPart](
        _compute_Q(data=data, graph=graph, target=('Qx', 'Qy'))
    )


providers = (
    sans_elastic,
    mask_wavelength_q,
    mask_wavelength_qxy,
    mask_and_scale_wavelength_q,
    mask_and_scale_wavelength_qxy,
    compute_Q,
    compute_Qxy,
)
